/*
 * $RCSfile: lcp_rom.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1994  Arizona Board of Regents
 *
 * $Log: lcp_rom.c,v $
 * Revision 1.2  1994/10/27 17:27:04  andrey
 * removed non-ansi definitions
 *
 * Revision 1.1  1994/10/24  20:46:02  andrey
 * Initial revision
 *
 *
 * $Revision: 1.2 $
 * $Date: 1994/10/27 17:27:04 $
 */

/*
 * lcp options processing routines
 */

#include "xkernel.h"
#include "romopt.h"
#include "lcp_constants.h"
#include "../ppp/ppp_fsm.h"
#include "../ppp/ppp_options.h"
#include "lcp_i.h"

#include "lcp_options.c"

static xkern_return_t   lcp_rom_term_ret( XObj, char **, int, int, VOID * );
static xkern_return_t   lcp_rom_conf_ret( XObj, char **, int, int, VOID * );
static xkern_return_t   lcp_rom_fail_ret( XObj, char **, int, int, VOID * );
static xkern_return_t   lcp_rom_timeout( XObj, char **, int, int, VOID * );
static xkern_return_t   lcp_rom_mru( XObj, char **, int, int, VOID * );
static xkern_return_t   lcp_rom_lqm_prot( XObj, char **, int, int, VOID * );
static xkern_return_t   lcp_rom_auth_prot( XObj, char **, int, int, VOID * );
static xkern_return_t   lcp_rom_pfc( XObj, char **, int, int, VOID * );
static xkern_return_t   lcp_rom_acfc( XObj, char **, int, int, VOID * );

xkern_return_t   lcpProcessOpts(XObj self);

extern int tracelcpp;

/*
 * lcp terminate_retries <num>
 * lcp configure_retries <num>
 * lcp failure_retries <num>
 * lcp timeout <num>
 * lcp mru <num>
 * lcp lqm_prot <num>
 * lcp auth_prot <num>
 * lcp prot_compression
 * lcp addr_prot_compression
 */

static XObjRomOpt       rOpts[] =
{
  {  "terminate_retries", 3, lcp_rom_term_ret },
  {  "configure_retries", 3, lcp_rom_conf_ret },
  {  "failure_retries", 3, lcp_rom_fail_ret },
  {  "timeout", 3, lcp_rom_timeout },
  {  "mru", 3, lcp_rom_mru },
  {  "lqm_prot", 3, lcp_rom_lqm_prot },
  {  "auth_prot", 3, lcp_rom_auth_prot },
  {  "prot_compression", 2, lcp_rom_pfc },
  {  "addr_prot_compression", 2, lcp_rom_acfc }
};

xkern_return_t
lcpProcessOpts(XObj self)
{

 /* XXXXXXXXXXXXXXXXX must add magic number ! XXXXXXXXXXXXXXXX */


  return findXObjRomOpts(self, rOpts, sizeof(rOpts)/sizeof(XObjRomOpt), 0);
}

static xkern_return_t
lcp_rom_term_ret(
    XObj        self,
    char        **strs,
    int         nFields,
    int		line,
    VOID        *arg)
{
  PSTATE * pstate = (PSTATE *)self->state;
  int k;

  if (sscanf(strs[2], "%d", &k) != 1) {
    xTraceP2(self, TR_SOFT_ERRORS, "can't read %s entry: %s", strs[1], strs[2]);
    return XK_FAILURE;
  }
  pstate->implem_options.term_retry = k;
  return XK_SUCCESS;
}

static xkern_return_t
lcp_rom_conf_ret(
    XObj        self,
    char        **strs,
    int         nFields,
    int         line,
    VOID        *arg)
{
  PSTATE * pstate = (PSTATE *)self->state;
  int k;

  if (sscanf(strs[2], "%d", &k) != 1) {
    xTraceP2(self, TR_SOFT_ERRORS, "can't read %s entry: %s", strs[1], strs[2]);
    return XK_FAILURE;
  }
  pstate->implem_options.config_retry = k;
  return XK_SUCCESS;
}

static xkern_return_t
lcp_rom_fail_ret(
    XObj        self,
    char        **strs,
    int         nFields,
    int         line,
    VOID        *arg)
{
  PSTATE * pstate = (PSTATE *)self->state;
  int k;

  if (sscanf(strs[2], "%d", &k) != 1) {
    xTraceP2(self, TR_SOFT_ERRORS, "can't read %s entry: %s", strs[1], strs[2]);
    return XK_FAILURE;
  }
  pstate->implem_options.config_naks = k;
  return XK_SUCCESS;
}

static xkern_return_t
lcp_rom_timeout(
    XObj        self,
    char        **strs,
    int         nFields,
    int         line,
    VOID        *arg)
{
  PSTATE * pstate = (PSTATE *)self->state;
  int k;

  if (sscanf(strs[2], "%d", &k) != 1) {
    xTraceP2(self, TR_SOFT_ERRORS, "can't read %s entry: %s", strs[1], strs[2]);
    return XK_FAILURE;
  }
  pstate->implem_options.timeval = k;
  return XK_SUCCESS;
}


static xkern_return_t
lcp_rom_mru(
    XObj        self,
    char        **strs,
    int         nFields,
    int         line,
    VOID        *arg)
{
  PSTATE * pstate = (PSTATE *)self->state;
  int k;
  MRU_opt * o;

  if (sscanf(strs[2], "%i", &k) != 1) {
    xTraceP2(self, TR_SOFT_ERRORS, "can't read %s entry: %s", strs[1], strs[2]);
    return XK_FAILURE;
  }
  o = X_NEW(MRU_opt);
  o->mru = k;
  ppp_add_option(&pstate->rom_lcp_options, o, LCP_MRU);
  return XK_SUCCESS;
}

static xkern_return_t
lcp_rom_lqm_prot(
    XObj        self,
    char        **strs,
    int         nFields,
    int         line,
    VOID        *arg)
{
  PSTATE * pstate = (PSTATE *)self->state;
  int k;
  QUAL_opt * o;

  if (sscanf(strs[2], "%i", &k) != 1) {
    xTraceP2(self, TR_SOFT_ERRORS, "can't read %s entry: %s", strs[1], strs[2]);
    return XK_FAILURE;
  }
  o = X_NEW(QUAL_opt);
  o->quality_prot = k;
  ppp_add_option(&pstate->rom_lcp_options, o, LCP_QUAL);
  return XK_SUCCESS;
}

static xkern_return_t
lcp_rom_auth_prot(
    XObj        self,
    char        **strs,
    int         nFields,
    int         line,
    VOID        *arg)
{
  PSTATE * pstate = (PSTATE *)self->state;
  int k;
  AUTH_opt * o;

  if (sscanf(strs[2], "%i", &k) != 1) {
    xTraceP2(self, TR_SOFT_ERRORS, "can't read %s entry: %s", strs[1], strs[2]);
    return XK_FAILURE;
  }
  o = X_NEW(AUTH_opt);
  o->auth_prot = k;
  ppp_add_option(&pstate->rom_lcp_options, o, LCP_AUTH);
  return XK_SUCCESS;
}

static xkern_return_t
lcp_rom_pfc(
    XObj        self,
    char        **strs,
    int         nFields,
    int         line,
    VOID        *arg)
{
  PSTATE * pstate = (PSTATE *)self->state;
  ppp_add_option(&pstate->rom_lcp_options, 0, LCP_PFC);
  return XK_SUCCESS;
}

static xkern_return_t
lcp_rom_acfc(
    XObj        self,
    char        **strs,
    int         nFields,
    int         line,
    VOID        *arg)
{
  PSTATE * pstate = (PSTATE *)self->state;
  ppp_add_option(&pstate->rom_lcp_options, 0, LCP_ACFC);
  return XK_SUCCESS;
 }
