/* 
 * ppp.c
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993 Bryce Nutter
 *
 *
 */

#include "ppp.h"

int The_State = Closed;

XObj      Pup;

void clear_buf(u_char *b_ptr,int size)
{
  u_char *ptr=b_ptr;
  int i;

  for(i=0;i < size; i++)
    *ptr++ = NULL;

}


int Free_blck(u_char *put_back)
{
    blck        *bp;
    int         found = FALSE,i;

    xTrace0(pppp, TR_FULL_TRACE, "PPP - Entering Free_blck");
    
    bp = &(pol.blocks[0]);

    for (i=0; i<MAX_SERIAL_BLOCKS && !found; i++) {
      if (bp->data == put_back) {
	  found = TRUE;
	  bp->ref = 0;
	  clear_buf((u_char *)bp->data,2*(DEVICE_MTU+TLEN+HLEN));
      }
      else
	bp = &(pol.blocks[i + 1]);
    }
    if (!found)
      xTrace0(pppp, TR_ERRORS,
		"PPP - could not get a free block");
    xTrace2(pppp, TR_MAJOR_EVENTS, "PPP - Free_blck returns block %x (# %d)",bp,i);

    xTrace0(pppp, TR_FULL_TRACE, "PPP - Exiting Free_blck");
    return (found);
}


void
Init_blocks()
{
    int 	i;
    blck 	*bp;
    static int	initialized = 0;

    xTrace0(pppp, TR_FULL_TRACE, "PPP - Entering Init_blocks");
    initialized = 1;
    pol.total_blocks = MAX_SERIAL_BLOCKS;
    pol.next_block = 0;
    
    bp = &(pol.blocks[0]);
    
    for (i=0; i<MAX_SERIAL_BLOCKS; i++) {
  
      if (!(bp->data = (u_char *) xMalloc(2*(DEVICE_MTU+TLEN+HLEN))))
	Kabort("Malloc of ppp packets - Init_blocks"); 
      clear_buf((u_char *)bp->data,2*(DEVICE_MTU+TLEN+HLEN));
      bp->ref = 0;

        if (i != (MAX_SERIAL_BLOCKS - 1))
	  bp = &(pol.blocks[i + 1]);
    }
    xTrace0(pppp, TR_FULL_TRACE, "PPP - Exiting Init_blocks");
}

static u_char *
nextBlck()
{
    blck 	*bp;
    int		i;

    xTrace0(pppp, TR_FULL_TRACE, "PPP -Entering nextBlck");
    i = pol.next_block;
    do {
	bp = &(pol.blocks[i]);
	i++;
	i %= pol.total_blocks;
    } while ( bp->ref == 1 && i != pol.next_block );
    if ( bp->ref == INUSE ) {
	printf("ppp: nextBlck - all blcks are busy\n");
	return NULL;
    }
    xTrace2(pppp, TR_MAJOR_EVENTS, "PPP - nextBlck returns block %x (# %d)",bp,i);
    bp->ref = INUSE;
    pol.next_block = i;
    xTrace0(pppp, TR_FULL_TRACE, "PPP - Exiting nextBlck");
    return bp->data;
}


int read_rom(pstate)
     PPPSTATE *pstate;
{
    int i;
    int num_ifs = 0;


    /*
     * Check the rom file for ppp initialization 
     */

    xTrace0(pppp, TR_FULL_TRACE, "PPP-Entering read_rom");
    num_ifs = pstate->n_interfaces;
    for ( i=0; rom[i][0]; i++ ) {
	if ( ! strcmp(rom[i][0], "ppp") ) {
	    if ( ! rom[i][1] || ! rom[i][2] ) {
		xError(sprintf(errBuf,
			       "PPP(in ip.c) ROM file format error in line %d",
			       i + 1 ));
		continue;
	    }

	    pstate->interface[num_ifs].MYPPPAddr_int = (int)inet_addr(rom[i][1]);
	    pstate->interface[num_ifs].THEIRPPPAddr_int = (int)inet_addr(rom[i][2]);

	    pstate->interface[num_ifs].mtu = atoi(rom[i][3]);
	    DEVICE_MTU = pstate->interface[num_ifs].mtu - TLEN - HLEN;
	    xTrace1(pppp, TR_EVENTS, "PPP - read_rom, mtu is: %d", DEVICE_MTU);
	    if ( !strcmp(rom[i][4], "ttyb") )
		pstate->interface[num_ifs].device = TTYB;
	    else
	      if ( !strcmp(rom[i][4], "ttya") )
		pstate->interface[num_ifs].device = TTYA;
	      else
		xError(sprintf(errBuf,
			       "PPP ROM file format error in line %d",i + 1 ));
	    
	    xTrace1(pppp,TR_MAJOR_EVENTS,"read_rom my ppp_addr %s ",
		  ipHostStr((IPhost *)&(pstate->interface[num_ifs].MYPPPAddr_int)));
	    xTrace1(pppp,TR_MAJOR_EVENTS,"read_rom their ppp_addr %s ",
		  ipHostStr((IPhost *)&(pstate->interface[num_ifs].THEIRPPPAddr_int)));
	    
	    num_ifs++;
	  }
      }
    xTrace0(pppp, TR_FULL_TRACE, "PPP -Exiting read_rom");
    return(num_ifs);
}

bool msg_peeky(ptr,len,arg)
     u_char *ptr;
     long len;
     VOID *arg;
{
  long count = len;
  u_char *p;
  int i=0,k=0;
  xTrace0(pppp, TR_FULL_TRACE, "PPP -Entering msg_peeky");
  p = ptr;
  while(k++ < count)
    printf("|%x",p[i++]);
  xTrace0(pppp, TR_FULL_TRACE, "PPP -Exiting msg_peeky");

  return(1);
}

/*
 * pppHdrStore -- writeheader to potentially unaligned msg buffer.
 * Note:  *hdr will be modified
 */
static void
pppHdrStore(hdr, dst, len, arg)
    VOID *hdr;
    u_char *dst;
    long int len;
    VOID *arg;
{
    xTrace0(pppp, TR_FULL_TRACE, "PPP -Entering  pppHdrStore");
    (u_char)*dst = Flag_h;dst++;
    (u_char)*dst = Address;dst++;
    (u_char)*dst = Control;dst++;
    bcopy((char*)&Protocolh,(char*)dst,2);
    xTrace0(pppp, TR_FULL_TRACE, "PPP -Exiting pppHdrStore");
}


/*
 * pppHdrLoad -- load header from potentially unaligned msg buffer.
 */
static long
pppHdrLoad(hdr, src, len, arg)
    VOID *hdr;
    u_char *src;
    long int len;
    VOID *arg;
{
  xTrace0(pppp, TR_FULL_TRACE, "PPP - Entering pppHdrLoad");

  Flag_h = (u_char)*src;src++;
  Address = (u_char)*src;src++;
  Control = (u_char)*src;src++;
  bcopy((char *)src,(char *)&Protocolh,2);

  xTrace0(pppp, TR_FULL_TRACE, "PPP -Exiting pppHdrLoad");
  return HLEN;
}


/*
 * ppp_init
 */
int
ppp_init(self)
    XObj self;
{
    Part	part;
    PPPSTATE	*pstate;
    int         i;
    XObj        llp;

    xTrace0(pppp, TR_EVENTS, "PPP - Entering ppp_init");
    xAssert(xIsProtocol(self));

    if ( ! xIsProtocol(llp = xGetDown(self, 0)) ) {
	xError("ppp can not get driver protocol object");
	return 0;
    }
    if ( xOpenEnable(self, self, llp, 0) == XK_FAILURE ) {
	xError("ppp cannot openenable driver protocol");
	return 0;
    }

    for(i=0;i<MAX_T_INFO;i++)
      {
	T_out_data[i].s = NULL;
	T_out_data[i].Code = 0x0;
	T_out_data[i].alloc = 0;
      }
    semInit(&ppp_enable,0);
    semInit(&ppp_disable,0);
    semInit(&link_sync,1);
    semInit(&demux_sem,1);
    semInit(&handle_sync,1);
    semInit(&device_lock,1);
    
    getproc_protl(self);

    pstate = X_NEW(PPPSTATE);
    self->state = (VOID *)pstate;
    sent_first = 0;
    received_first = 0;
    receive_mode = 0;

    pstate->activemap = mapCreate(ACTIVE_MAP_SIZE, sizeof(ActiveId));
    pstate->passivemap = mapCreate(PASSIVE_MAP_SIZE, sizeof(PassiveId)); 

    partInit(&part, 1);

    partPush(part, ANY_HOST,sizeof(Part));
    ppp_obj = self;

    pstate->n_interfaces = 0;
    read_rom(pstate);
    xTrace0(pppp, TR_EVENTS, "PPP - Exiting ppp_init");
    return 0;
}

/*
 * pppOpen
 */
static XObj
pppOpen( self, hlpRcv, hlpType, p )
    XObj self, hlpRcv, hlpType;
    Part *p;
{
    XObj	ppp_s;
    XObj        ppp = xMyProtl(self);
    XObj        llp = xGetDown(ppp,0);
    XObj        lls;
    xkern_return_t ppp_s_ret;
    ActiveId 	key;
    PPPSTATE	*pstate;
    LCPHDR      lcp_h_r;
    CONF_OPT    co;
    long        lcp_len;
    u_short     mru;
    u_char      *lcp_buf_r;
    Msg         lcp_msg_r;
    int         num_ifs=0;

    xTrace0(pppp, TR_EVENTS, "PPP - Entering pppOpen");
    pstate = (PPPSTATE *)self->state;
    key.device_name = pstate->interface[num_ifs].device;
    wus = 1;

    xTrace1(pppp, TR_MAJOR_EVENTS, "PPP open: to device %d (in ppp)",
	    key.device_name);
      
    partInit(p, 1);
    partPush(*p, &key.device_name,sizeof(int));
    if ( (lls = xOpen(self, self, llp, p)) == ERR_XOBJ ) 
	xTrace0(pppp, TR_EVENTS,
		"pppOpen  could not get direct lower session");
    xAssert(xIsSession(lls));

    ppp_s_ret = mapResolve(pstate->activemap,(char*)&key,(void **)&ppp_s);

    if(ppp_s_ret != XK_SUCCESS) {
	ppp_s = pppCreateSessn(self,hlpRcv,&key);
	if (The_State != Opened) {
	    The_State = Closed;
	    ((SSTATE*)(ppp_s->state))->Protocol = LCP;
	    ((SSTATE*)(ppp_s->state))->State = Closed;
	    mru = pstate->interface[0].mtu;

	    xTrace1(pppp, TR_MAJOR_EVENTS, "PPP open: sending mtu: %d",
	    mru);
	    lcp_len = LLEN + COLEN;
	    msgConstructEmpty(&lcp_msg_r);
	    msgConstructAllocate(&lcp_msg_r,lcp_len,(char **)&lcp_buf_r);
	    lcp_h_r.Code = CONF_REQ;
	    lcp_h_r.Identifier = 0xf;
	    lcp_h_r.Length = LLEN;  

	    co.Type = MRU;
	    co.Length = MRU_FORMAT_LEN;
	    co.Mru = mru;

	    bcopy((char *)&lcp_h_r,lcp_buf_r,LLEN);
	    lcp_buf_r += LLEN;
	    bcopy((char *)&co,lcp_buf_r,COLEN);

	    xTrace1(pppp, TR_MAJOR_EVENTS, "PPP open: pushing to session: %d",ppp_s);
	    xPush(ppp_s,&lcp_msg_r);
	    The_State = Req_Sent;
	    semWait(&ppp_enable);
	    msgDestroy(&lcp_msg_r);
	  }
      }
    else {
	((SSTATE*)(ppp_s->state))->Protocol = IP;
	((SSTATE*)(ppp_s->state))->State = The_State;
      }

    xTrace0(pppp, TR_EVENTS, "PPP - Exiting pppOpen");
    return (ppp_s);
}


static XObj
pppCreateSessn(self, hlp, key)
    XObj self;
    XObj hlp;
    ActiveId *key;
{
    XObj	s,s_ret;
    SSTATE	*sstate;
    PPPSTATE	*pstate;

    xTrace0(pppp, TR_EVENTS, "PPP - Entering pppCreateSessn");
    pstate = (PPPSTATE *)self->state;
    s = xCreateSessn(getproc_sessn, hlp, self,self,0,0);
    xAssert(xIsSession(s));

    s->binding = (Bind)mapBind(pstate->activemap, (char *)key, (int)s);

    if ( s->binding == ERR_BIND ) {
	xTrace0(pppp, TR_EVENTS, "PPP - pppCreatesessn - session already existed");
	pppCloseSessn(s);
	mapResolve(pstate->activemap, (char *)key, (void **)&s_ret);
	return s_ret; 
    }

    if (!(sstate = (SSTATE *)xMalloc(sizeof(SSTATE))))
      Kabort("Malloc of ppp packets-pppCreateSessn"); 
    xTrace0(pppp, TR_MAJOR_EVENTS,
		"PPP create sessn ");
    sstate->Protocol = IP;  /* PPP */
    sstate->State = Listen; 

    s->state = (char *)sstate;
    xTrace0(pppp, TR_EVENTS, "PPP - Exiting pppCreateSessn");
    return s;
}


/*
 * pppControlSessn
 */
int
pppControlSessn(s, opcode, buf, len)
    XObj s;
    int opcode;
    char *buf;
    int len;
{
  SSTATE        *sstate;
  PPPSTATE        *pstate;
  int c;

  xTrace0(pppp, TR_EVENTS, "PPP ControlSessn");
  
  xAssert(xIsSession(s));
  
  sstate = (SSTATE *) s->state;
  pstate = (PPPSTATE *) s->myprotl->state;
  
  switch (opcode) {
    
  case GETMYPROTO:
    checkLen(len, sizeof(long));
    return sizeof(long);

  case GETPEERPROTO:
    checkLen(len, sizeof(long));
    return sizeof(long);

   case GETPARTICIPANTS:
     {
	 Part	*p = (Part *)buf;
	 int 	retLen;
	 long	*key;

	 retLen = xControl(xGetDown(s, 0), opcode, buf, len);
	 if ( retLen >= sizeof(Part) ) {
	     key = X_NEW(long);
	     partPush(p[0], key, sizeof(Part));
	 }
	 if ( retLen >= 2 * sizeof(Part) ) {
	     key = X_NEW(long);
	     partPush(p[0], key, sizeof(Part));
	 }
	 return retLen;
     }

  case GETMAXPACKET:
    c = (pstate->interface[0].mtu - TLEN - HLEN);
    *(int*)buf = c;
    return c;
  case GETOPTPACKET:
    {
	checkLen(len, sizeof(int));
	if ( xControl(xGetDown(s, 0), opcode, buf, len) < sizeof(int) ) {
	    return -1;
	}
	 *(int *)buf -= (HLEN + TLEN);
	return sizeof(int);
    }
    
  default:
    return xControl(xGetDown(s, 0), opcode, buf, len);;
  }
}


/*
 * pppControlProtl
 */
int
pppControlProtl(self, opcode, buf, len)
    XObj self;
    int opcode;
    char *buf;
    int len;
{
  PPPSTATE	*ps = (PPPSTATE *)self->state;
  int 	reply,num_ifs=0;
  ArpBinding	*b = (ArpBinding *)buf;
 
  xTrace0(pppp, TR_EVENTS, "PPP - Entering ControlProtl");
  xAssert(xIsProtocol(self));
    
    switch (opcode) {
	
    case GETMAXPACKET:
      reply = (ps->interface[num_ifs].mtu - TLEN - HLEN);
      break;

    case RESOLVE:
      checkLen(len, sizeof(ArpBinding));
      if ( !strcmp((char *)ipHostStr((IPhost*)&b->ip),
		   (char *)ipHostStr((IPhost*)&ps->interface[num_ifs].THEIRPPPAddr_int)) )
	  reply = sizeof(ArpBinding);
      else
	reply = -1;
      break;

    case ARP_GETMYBINDING:
      checkLen(len, sizeof(ArpBinding));
      b->ip = *(IPhost*)&ps->interface[num_ifs].MYPPPAddr_int;
      reply = sizeof(ArpBinding);
      break;
      
    default:
      reply = -1;
	
    }
  xTrace0(pppp, TR_EVENTS, "PPP - Exiting ControlProtl");
  return(reply);
}


/*
 * pppDemux
 */
xkern_return_t
pppDemux(self,lls,msg)
    XObj   self;
    XObj   lls;
    Msg    *msg;
{
    XObj sself;
    Msg *dg;
    XObj   	s;
    ActiveId  	activeid;
    PassiveId	key;
    PPPSTATE      *pstate;
    u_char      bb[30];
    xkern_return_t e_ret,s_ret;

    Enable	*e;
    long        len,tail_len,lcp_len=0;
    LCPHDR      lcp_h,lcp_h_r;
    CONF_OPT    co;
    char        *lcp_buf,*lcp_buf_r;
    Msg         lcp_msg,lcp_msg_r,tail_data,ip_d_msg;
    int         no_pop = 0,k,n,num_opts=0,opt_ct=0;
    u_char     *buffer1,*buffer4;

    semWait(&demux_sem);

    dg = msg;
    sself = self;

    bb[1] = NULL;
    k = msgLen(msg);
    pstate = (PPPSTATE *)sself->state;
    xTrace0(pppp, TR_EVENTS, "PPP Demux (in ppp)");

    len = k;
    msgPop(dg, pppHdrLoad, 0, HLEN, dg);

    n = k-HLEN;
    s_ret = mapResolve(pstate->activemap, (char *) &activeid,&s);
    if (s_ret == XK_FAILURE) {
      key.device_name = pstate->interface[0].device;
	e_ret = mapResolve(pstate->passivemap, (char *) &key,(void **)&e);
	if (e_ret == XK_FAILURE) {
	    xTrace0(pppp, TR_MAJOR_EVENTS, "pppDemux dropping the message (in ppp)");
	    return XK_SUCCESS;
	}
	xTrace1(pppp, TR_MAJOR_EVENTS, "Found an open enable for prot %d (in ppp)",
		e->hlpRcv);
	s = pppCreateSessn(sself, e->hlpRcv, &activeid);
	if (s == ERR_XOBJ) {
	    xTrace0(pppp, TR_ERRORS, "pppDemux could not create session (in ppp)");
	    return XK_SUCCESS;
	}
      xOpenDone(e->hlpType,s,sself);
    } else {
	xTrace1(pppp, TR_EVENTS, "Popping to existing session %x (in ppp)", s);
      }
    xAssert(xIsSession(s));
    tail_len = TLEN;
    k -= HLEN;
    len -= HLEN;
    xTrace1(pppp, TR_EVENTS, "PPPdemux len is %d",len);
    xTrace1(pppp, TR_EVENTS, "PPPdemux k(len-HLEN) is %d",k);

    msgConstructAllocate(&ip_d_msg,k-TLEN,(char **)&buffer1);
    msgConstructAllocate(&tail_data,k,(char **)&buffer4);

    len = len - TLEN;
    msgChopOff(dg,&tail_data,k);

    k -= TLEN;
    msgChopOff(&tail_data,dg,k);
    n = TLEN;

    msgPop(&tail_data,pppTailLoad,&tail_data,TLEN,&tail_data);
    msgDestroy(&tail_data);
    msgDestroy(&ip_d_msg);

    if (Protocolh == LCP)
      {
	no_pop = 1;

	msgConstructEmpty(&lcp_msg);
	msgConstructAllocate(&lcp_msg,msgLen(dg),&lcp_buf);
	msgConstructCopy(&lcp_msg,dg);
	lcp_len = LLEN;
	msgForEach(dg,get_data,lcp_buf);

	bcopy(lcp_buf,(char *)&lcp_h,LLEN);
	num_opts = (msgLen(dg) - LLEN)/COLEN;

	lcp_len = LLEN;
	msgConstructEmpty(&lcp_msg_r);
	msgConstructAllocate(&lcp_msg_r,lcp_len,&lcp_buf_r);
	lcp_h_r.Identifier = 0xf;
	lcp_h_r.Length = LLEN;

	switch (The_State) {
	  case Listen:
	    if (lcp_h.Code == CONF_REQ) {
		((SSTATE*)(s->state))->State = Ack_Sent;

		lcp_h_r.Code = CONF_ACK;

		if (num_opts >= 1)
		  do {
		    lcp_buf += LLEN;
		    bcopy(lcp_buf,(char *)&co,COLEN);
		    opt_ct += 1;
		    switch (co.Type) {
		    case MRU:
		      if (co.Mru != pstate->interface[0].mtu) {
			xTrace0(pppp, TR_EVENTS,"PPP: WARNING-MTU's do not match");
		        xTrace1(pppp, TR_EVENTS,"Receiving MRU: %d",co.Mru);
			xTrace1(pppp, TR_EVENTS,"My MRU: %d",pstate->interface[0].mtu);
		      }
		      break;
		    case ACCM:
		    case AP:
		    case QP:
		    case MN:
		    case PFC:
		    case ACFC:
		      xTrace0(pppp, TR_EVENTS, "PPP:Warning-Unsupported Option.");
		    default:
		      xTrace0(pppp, TR_EVENTS, "PPP:Warning-Option not recognized.");
		      break;
		    }
		  } while(opt_ct < num_opts);

		bcopy((char *)&lcp_h_r,lcp_buf_r,LLEN);
		((SSTATE*)(s->state))->Protocol = LCP;
		xPush(s,&lcp_msg_r);
		Delay(5);

		msgDestroy(&lcp_msg_r);
		msgConstructEmpty(&lcp_msg_r);
		msgConstructAllocate(&lcp_msg_r,lcp_len,&lcp_buf_r);
		lcp_h_r.Code = CONF_REQ;
		lcp_h_r.Identifier = 0xf;
		lcp_h_r.Length = LLEN;  
		bcopy((char *)&lcp_h_r,lcp_buf_r,LLEN);
		((SSTATE*)(s->state))->Protocol = LCP;
		xPush(s,&lcp_msg_r);
		The_State = Ack_Sent;
	      }
	    else
	      xTrace0(pppp, TR_EVENTS, "DID NOT FIND CONF_REQ in ppp_demux");
	    
	    break;
	  case Req_Sent:
	    if (lcp_h.Code == CONF_REQ) {
		((SSTATE*)(s->state))->State = Ack_Sent;
		lcp_h_r.Code = CONF_ACK;
		bcopy((char *)&lcp_h_r,lcp_buf_r,LLEN);
		((SSTATE*)(s->state))->Protocol = LCP;
		xPush(s,&lcp_msg_r);
		The_State = Ack_Sent;
	      }
	    else 
	      if (lcp_h.Code == CONF_ACK) {
		  ((SSTATE*)(s->state))->State = Ack_Rcvd;
		  The_State = Ack_Rcvd;
		}
	      else
		xTrace0(pppp, TR_EVENTS, "DID NOT FIND CONF_REQ in ppp_demux");  
	    break;
	  case Ack_Rcvd:
	    if (lcp_h.Code == CONF_REQ) {
		lcp_h_r.Code = CONF_ACK;
		bcopy((char *)&lcp_h_r,lcp_buf_r,LLEN);
		((SSTATE*)(s->state))->Protocol = LCP;
		xPush(s,&lcp_msg_r);
		The_State = Opened;
		semSignal(&ppp_enable);
	      }
	    else
	      if (lcp_h.Code == CONF_ACK) {
		  ((SSTATE*)(s->state))->State = Ack_Rcvd;
		  The_State = Req_Sent;
		  lcp_h_r.Code = CONF_REQ;
		  bcopy((char *)&lcp_h_r,lcp_buf_r,LLEN);
		  ((SSTATE*)(s->state))->Protocol = LCP;
		  xPush(s,&lcp_msg_r);
		}
	      else
		xTrace0(pppp, TR_EVENTS, "DID NOT FIND CONF_REQ in ppp_demux");
	    break;
	  case Ack_Sent:
	    if (lcp_h.Code == CONF_REQ) {
		lcp_h_r.Code = CONF_ACK;
		bcopy((char *)&lcp_h_r,lcp_buf_r,LLEN);
		((SSTATE*)(s->state))->Protocol = LCP;
		xPush(s,&lcp_msg_r);
	      }
	    else
	      if (lcp_h.Code == CONF_ACK) {
		  ((SSTATE*)(s->state))->Protocol = IP;
		  semSignal(&ppp_enable);
		  The_State = Opened;
		}
	      else
		xTrace0(pppp, TR_EVENTS, "No conf_ack or conf_req in ppp_demux");
	    break;
	  case Opened:
	    xTrace0(pppp, TR_EVENTS, "STATE IS OPENED IN PPP_DEMUX");
	    break;
	  case Closed:
	    xTrace0(pppp, TR_EVENTS, "State is Closed in ppp_demux");
	    break;	    
	  default:
	    xTrace0(pppp, TR_EVENTS, "The_State is unknown in ppp_demux");
	    break;
	  }
      }
    if (!no_pop) {
	xAssert(xIsSession(s));
	okayed++;
	semSignal(&demux_sem);
	return xPop(s, 0, dg,0);
      }
    else if (Protocolh == LCP) {
      msgDestroy(&lcp_msg_r);
      msgDestroy(&lcp_msg);
    }
    semSignal(&demux_sem);
    xTrace0(pppp, TR_EVENTS, "Leaving PPPdemux");
    return XK_SUCCESS;     
  }


/*
 * pppOpenEnable
 */
static xkern_return_t
pppOpenEnable( self, hlpRcv, hlpType, p )
    XObj self, hlpRcv, hlpType;
    Part *p;
{
    PPPSTATE 	*pstate;
    Enable 	*e;
    xkern_return_t e_ret;
    PassiveId   key2;
    int         num_ifs=0;

    The_State = Listen;
    xTrace0(pppp, TR_EVENTS, "PPP - Entering pppOpenEnable");
    
    pstate = (PPPSTATE *)self->state;

    key2.device_name = pstate->interface[num_ifs].device;

    e_ret = mapResolve(pstate->passivemap, (char *)&key2, (void **)&e);
    if ( e_ret == XK_SUCCESS) {
	if ( e->hlpRcv == hlpRcv ) {
	    e->rcnt++;
	    return XK_SUCCESS;
	}
	return XK_FAILURE;
    }
    if (!(e = (Enable *)xMalloc(sizeof(Enable))))
      Kabort("Malloc of ppp packets-pppOpenEnable"); 
    e->hlpRcv = hlpRcv;
    Pup = hlpRcv;
    e->hlpType = hlpType;
    e->rcnt = 1;
    e->binding = mapBind(pstate->passivemap, (char *)&key2, (int) e);
    if (e->binding == ERR_BIND) {
      xFree((char *)e);
      return XK_FAILURE;
    }
    return XK_SUCCESS;
  }

static xkern_return_t
pppCloseSessn(s)

    XObj	s;
{
    PPPSTATE	*pstate;
    SSTATE	*sstate;

    xAssert(xIsSession(s));
    xTrace1(pppp, TR_MAJOR_EVENTS, "PPP CLOSING session(does zippo) %x (in ppp)", s);
    xAssert(s->rcnt == 0);
    pstate = (PPPSTATE *)s->myprotl->state;
    sstate = (SSTATE *)s->state;
    return XK_SUCCESS;
}


/*
 * pppCloseProtl
 */
static xkern_return_t
pppCloseProtl(self)
    XObj self;
{
  PPPSTATE        *pstate;
  
  xAssert(xIsProtocol(self));
  xTrace0(pppp, TR_MAJOR_EVENTS, "PPP - CloseProtl");
  if (((self->rcnt)) <= 0) {
      pstate = (PPPSTATE *) self->state;
      mapClose(pstate->activemap);
      mapClose(pstate->passivemap);
      xFree((char *) pstate);
  }
  return XK_SUCCESS;
}

void dumb_func(char* dp)
{
  xTrace0(pppp, TR_MAJOR_EVENTS, "PPP - dumb_func does nothing");
}

/*
 * pppPush
 */
static xmsg_handle_t
pppPush(s, msg)
    XObj s;
    Msg *msg;
{

  SSTATE        *sstate;
  XObj          lls;
  XObj          ppp;
  XObj          llp;
  int           k,j,b,i,r_b_ptr,ip_len,Sent;
  int           Done = 0,rval_total;
  float         rval_f;
  ActiveId 	key;
  Part          p;
  storeInfo	info;
  u16        Crc;
  u_char *bufPtr,*real_bufPtr,*o_bufPtr,*o_real_bufPtr;
  u_char *track,*buffer2,*buffer3;
  Msg   ip_msg,tail,f_msg;
  long          msg_len,new_msg_len;

  xTrace0(pppp, TR_EVENTS, "PPP - Entering pppPush");
  semWait(&device_lock);
  if (!(bufPtr = (u_char *) nextBlck()))
    Kabort("NextBlck of ppp packets-pppPush,A"); 
  o_bufPtr = bufPtr;

  if (!(real_bufPtr = (u_char *) nextBlck()))
    Kabort("NextBlck of ppp packets-pppPush,B"); 
  o_real_bufPtr = real_bufPtr;

  xTrace0(pppp, TR_EVENTS, "in ppp push");
  
  /*printf("PPP pushing message: ");
  msgForEach(msg,msg_peeky,0);
  printf("\n");*/

  xAssert(xIsSession(s));

  sstate = (SSTATE *) s->state;
  Flag_h = FLAG;
  Address = ADDRESS;
  Control = CONTROL;
  Protocolh = sstate->Protocol;
  info.s = s;
  info.m = msg;

  
  msg_len = msgLen(msg);

  ip_len = msg_len;

  switch (The_State) { 
  case Listen:
    break;
  case Req_Sent:
    break;
  case Ack_Rcvd:
    break;
  case Ack_Sent:
    break;
  case Opened:
    Protocolh = sstate->Protocol;
    break;
  case Closed:
    sstate->Protocol = IP;
    break;
  default:
    break;
  }

  msg_len = TLEN + msg_len;

  msgConstructAllocate(&ip_msg,msg_len,(char **)&buffer3);
  msgConstructAllocate(&tail,TLEN,(char **)&buffer2);

  xTrace1(pppp,TR_EVENTS,"msg len of ip datagram plus tail in ppp is: %d",(msg_len));

  Fcs = FCS_CHECK;
  Flag_t = FLAG;

  bcopy((char*)&Fcs,(char *)buffer2,2);
  buffer2++; buffer2++;
  *buffer2 = Flag_t;

  msgJoin(&ip_msg,msg,&tail);
  xTrace1(pppp, TR_EVENTS, "ppppush, length of message before ppphdr_push is %d",msgLen(&ip_msg));

  msgPush(&ip_msg, pppHdrStore,0, HLEN, &info);
  msg_len = msgLen(&ip_msg);
  k = msgLen(&ip_msg);

  big_len = 0;
  msgForEach(&ip_msg,msg2Buf,&bufPtr);

  Crc = PPPINITFCS;
      
  bufPtr++;
  k = msg_len - 4;
  Crc = pppfcs(Crc,bufPtr,k);

  bufPtr += (msg_len-4);

  bcopy((char*)&Crc,(char*)bufPtr,2);

  r_b_ptr=0;
  bufPtr = o_bufPtr;

  for(b=0;b<msg_len;b++) {
      if ( (b != 0) && (b != (ip_len + TLEN + HLEN - 1)) )
	{
	  if ((*bufPtr == FLAG) || 
	      (*bufPtr == CEO) ||
	      (*bufPtr < XOR)) {
	      *real_bufPtr++ = CEO;
	      r_b_ptr++;
	      *real_bufPtr++ = (*bufPtr ^ XOR);
	      r_b_ptr++;
	    }
	  else {
	      *real_bufPtr++ = *bufPtr;
	      r_b_ptr++;
	    }
	}
      else {
	  *real_bufPtr++ = *bufPtr;
	  r_b_ptr++;
	}
      bufPtr++;
    }
  msg_len = r_b_ptr;

  k = msg_len;

  track = o_real_bufPtr;
  j = 0;
  if((k = msg_len%CHUNK) != 0) {
      j = CHUNK;
      j -= k;
      for(i=0;i < j;i++)
	*real_bufPtr++ = ADDRESS;
    }

  new_msg_len = msg_len + j;

  if(sstate->Protocol == IP)
    if(prev_write != 0)
      rval_f = prev_write;
  Done = 0;
  Sent = 0;
  rval_total = 0;

  track = o_real_bufPtr;

  msgConstructInplace(&f_msg,(char *)o_real_bufPtr,(long)new_msg_len,dumb_func);

  if ((lls = xGetDown(s,0)) == ERR_XOBJ) {
    xTrace0(pppp, TR_EVENTS, "PPPpush - could not get lls");
    partInit(&p, 1);
    partPush(p, &key.device_name,sizeof(int));
    ppp = xMyProtl(ppp_obj);
    llp = xGetDown(ppp,0);
    if ( (lls = xOpen(ppp,ppp, llp, &p)) == ERR_XOBJ ) 
	xTrace0(pppp, TR_EVENTS,
		"pppOpen  could not get direct lower session");
  }
    

  xTrace1(pppp, TR_EVENTS, "PPPpush - about to write %d to device driver",new_msg_len);
  xAssert(xIsSession(lls));

  xPush(lls,&f_msg);

  xTrace1(pppp, TR_EVENTS, "PPPpush - just wrote %d bytes to device driver",new_msg_len);

  prev_write = rval_total;
  sstate->Protocol = IP; 
  if (!(Free_blck((u_char*)o_bufPtr)))
    xTrace0(pppp, TR_MAJOR_EVENTS, "PPP problem calling Free_blck");

  if (!(Free_blck((u_char*)o_real_bufPtr)))
    xTrace0(pppp, TR_MAJOR_EVENTS, "PPP problem calling Free_blck");

  msgDestroy(&f_msg);
  msgDestroy(&ip_msg);
  msgDestroy(&tail);

  semSignal(&device_lock);

  xTrace0(pppp, TR_MAJOR_EVENTS, "PPP leaving pppPush");
  return XMSG_NULL_HANDLE;
}

static bool
msg2Buf(u_char *msgPtr, long len, void *bufPtr)
{
  int b_len = big_len,length=0;

   do {
       (*(u_char **)bufPtr)[b_len++] = (msgPtr)[length++];
   } while(length < len);
  big_len = b_len;

  return TRUE;
}

bool get_data(ptr,len,arg)
     char *ptr;
     long len;
     VOID *arg;
{
  char *p;

  p = ptr;
  bcopy(p,(char *)arg,len);
  return(1);
}

static long
pppTailLoad(hdr, src, len, arg)
    VOID *hdr;
    u_char *src;
    long int len;
    VOID *arg;
{
  xTrace0(pppp, TR_FULL_TRACE, "PPP TailLoad");
  bcopy((char *)src,(char *)&Fcs,2);
  src++;src++;
  Flag_t = (u_char)*src;
  return TLEN;
}


/*
 * pppPop
 */
static xkern_return_t
pppPop(s, ds, dg, h)
    XObj s;
    XObj ds;
    Msg *dg;
    VOID *h;
{

  xTrace1(pppp, TR_EVENTS, "PPP pop, length = %d (in ppp)", msgLen(dg));
  xAssert(xIsSession(s));

  /*printf("PPP popping message: ");
  msgForEach(dg,msg_peeky,h);
  printf("\n");*/

  return xDemux(s, dg);
}


static xkern_return_t
pppOpenDisable(self, hlpRcv, hlpType, p)
    XObj self, hlpRcv, hlpType;
    Part *p;
{
    PassiveId	key;
    PPPSTATE   	*pstate;
    Enable	*e;
    xkern_return_t e_ret;
    
    xTrace0(pppp, TR_MAJOR_EVENTS, "PPP open disable");
    pstate = (PPPSTATE *)self->state;

    xTrace1(pppp, TR_MAJOR_EVENTS, "key %d (in ppp)", key);
    e_ret = mapResolve(pstate->passivemap, (char *)&key, (void **)&e);
    if ( e_ret == XK_SUCCESS && e->hlpRcv == hlpRcv && e->hlpType == hlpType ) {
	if (--(e->rcnt) == 0) {
	    mapRemoveBinding(pstate->passivemap, e->binding);
	    xFree((char *)e);
	}
	return XK_SUCCESS;
    } else
	return XK_FAILURE;
}


static void
getproc_protl(s)
    XObj s;
{
  xTrace0(pppp, TR_EVENTS, "PPP getproc_protl");
  xAssert(xIsProtocol(s));
  s->close = pppCloseProtl;
  s->control = pppControlProtl;
  s->open = pppOpen;
  s->openenable = pppOpenEnable;
  s->opendisable = pppOpenDisable;
  s->demux = pppDemux;
}


static void
getproc_sessn(s)
    XObj s;
{
  xTrace0(pppp, TR_EVENTS, "PPP getproc_sessn");
  s->push = pppPush;
  s->pop = pppPop;
  s->control = pppControlSessn;
  s->close = pppCloseSessn;
}

