/*
 * $RCSfile: ppp_internal.h,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1994  Arizona Board of Regents
 *
 * $Log: ppp_internal.h,v $
 * Revision 1.1  1994/10/24 20:44:34  andrey
 * Initial revision
 *
 *
 * $Revision: 1.1 $
 * $Date: 1994/10/24 20:44:34 $
 */

/*
 *  miscellaneous protocol internal state
 */

typedef enum { PPP_dead,	/* initial state */
               PPP_estab,	/* LCP option negotiation occurs */
               PPP_auth,	/* CHAP or PAP authentication here */
               PPP_net,		/* NCP runs here, and upper protocol */
               PPP_open,	/* not per rfc, but PPP_term is fleeting */
                                /* so this indicates link is ready/working */
               PPP_term		/* exchanges terminate packets if possible */
             } ppp_state;

/*
 * paths through phases
 *
 * normal path
 *   PPP_dead -> PPP_estab -> PPP_auth -> PPP_net ->
 *               PPP_open -> PPP_term -> PPP_dead
 *
 * configuration negotiation failure
 *   PPP_dead -> PPP_estab -> PPP_dead
 *
 * authentication failure
 *   PPP_dead -> PPP_estab -> PPP_auth -> PPP_term -> PPP_dead
 *
 */

#define	PPP_NUMINTERFACES		10
#define ACTIVE_MAP_SIZE			10
#define PASSIVE_MAP_SIZE		10
#define	CONTROL_MAP_SIZE		30

typedef u_short ppp_protnum;

static ppp_protnum lookup_control_protnum(ppp_protnum);

typedef struct
{
  IPhost local;
  IPhost remote;
  char * devname;
  u_short mtu;

  /* debugging */
  u_short port;
  /* other options? */
} IP_LINK;

#define	IP_LINK_init(IPL_ptr)	()

/*
\
	((IPL_ptr)->local = 0,(IPL_ptr)->remote = 0,\
	 (IPL_ptr)->devname = 0, (IPL_ptr)->mtu = PPP_MAXPACKET,\
	 (IPL_ptr)->port = 0 )
 */

typedef struct
{
  Map		Active_map;
  Map		Passive_map;
  /*
   * maps protocol ids to control protocol sessions of same type
   */
  Map		Control_Prot_map;
  Map		Control_Sess_map;
  IP_LINK	IP_link;
  ppp_state	phase;
  XObj		LCP_session;		/* when LCP is up, copy it here */
  XObj		Out_session;		/* session down and of here */

  XObj		control_waiting;	/* control sess waiting negotiation */

} PSTATE;

#define	PSTATE_init(PS_ptr) \
  (\
   (PS_ptr)->Active_map = mapCreate(ACTIVE_MAP_SIZE, sizeof(ActiveID)),\
   (PS_ptr)->Passive_map = mapCreate(PASSIVE_MAP_SIZE, sizeof(PassiveID)),\
   (PS_ptr)->Control_Prot_map = mapCreate(CONTROL_MAP_SIZE, \
                                          sizeof(ppp_protnum)), \
   (PS_ptr)->Control_Sess_map = mapCreate(CONTROL_MAP_SIZE, \
                                          sizeof(ppp_protnum)), \
/*   (PS_ptr)->phase = PPP_dead, IP_LINK_init(&(PS_ptr)->IP_link),\ */ \
   (PS_ptr)->LCP_session = ERR_XOBJ,\
   (PS_ptr)->Out_session = ERR_XOBJ,\
   (PS_ptr)->control_waiting = ERR_XOBJ\
  )



typedef struct
{
  ppp_protnum 	prot;		/* protocol type */
} SSTATE;

#define	SSTATE_init(ST_ptr)	((ST_ptr)->prot = 0)

typedef struct
{
  ppp_protnum prot;
} ActiveID;

typedef struct
{
  ppp_protnum prot;
} PassiveID;

