#include "xkernel.h"
#include "prot/join.h"
#include "prot/crypt.h"
#include "prot/kerb_serv_bottom.h"
#include <krb5/krb5.h>
#include <isode/psap.h>
#include <krb5/kdb.h>
#include <krb5/asn1.h>
#include <com_err.h>
#include <krb5/asn.1/encode.h>
#include "prot/recvauth.h"


#ifdef __STDC__

void                  recvauth_init( XObj );
static xkern_return_t recvauthOpenDone( XObj, XObj, XObj, XObj );
static xkern_return_t recvauthOpenEnable( XObj, XObj, XObj, Part * );
static xkern_return_t recvauthDemux( XObj, XObj, Msg * );
static xkern_return_t recvauthPop( XObj, XObj, Msg *, void * );
static void           recvauth_sessn_init( XObj );
static int            recvauthProtlControl( XObj, int, char *, int );
static int            recvauthSessnControl( XObj, int, char *, int );
static int            process_ap_request( XObj );
static void           mycopypush( void *, char *, long, void * );
static long           mycopypop( void *, char *, long , void * );
static int            partToId( Part *, char *, int );
static int            part2ToId( Part*, char *, int );

#else

void                  recvauth_init();
static xkern_return_t recvauthOpenDone();
static xkern_return_t recvauthOpenEnable();
static xkern_return_t recvauthDemux();
static xkern_return_t recvauthPop();
static void           recvauth_sessn_init();
static int            recvauthProtlControl();
static int            recvauthSessnControl();
static int            process_ap_request();
static void           mycopypush();
static long           mycopypop();
static int            partToId();
static int            part2ToId();


#endif

typedef struct SSTATEstruct {
  XObj lower[4];
  int  curr;
  Bind tag_bind;
  Msg  *m[4];
  krb5_enc_tkt_part *enc_tkt_part;
} SSTATE;

typedef struct PSTATEstruct {
  Map activeMap;
  Map tagMap;
  Map pasMap;
  int paslen;
  int taglen;
  int status;
} PSTATE;


int tracerecvauthp = 1;
#define RECVP recvauthp
#define MAP_SZ 32


void
recvauth_init(self)
     XObj self;
{
  Part            p[1];
  long            port=4333;
  int             i;
  xkern_return_t  retval;
  IPhost myHost;
  PSTATE *ps;

  xTrace0(RECVP, TR_MAJOR_EVENTS, "recvauth_init entered.");

  self->state = xMalloc(sizeof(PSTATE));
  if (!self->state) {
    Kabort("malloc failed for state...exiting.");
  }
  bzero((char*)self->state, sizeof(PSTATE));
  ps = (PSTATE *)self->state;
  ps->activeMap = mapCreate(20, sizeof(XObj));
  ps->tagMap = mapCreate(20, 100);

  self->demux = recvauthDemux;
  self->opendone = recvauthOpenDone;
  self->control = recvauthProtlControl;
  self->openenable = recvauthOpenEnable;

#if 0
  xControl(xGetDown(self, 0), GETMYHOST, (char *)&myHost, sizeof(IPhost));

  for (i=0; i < 4; i++)
    {
      partInit(p,1);
      partPush(p[0], &myHost, sizeof(IPhost));
      partPush(p[0], (void *) &port, sizeof(long));

      retval = xOpenEnable(self, self, xGetDown(self,i), p);
      if (retval == XK_FAILURE) {
        xTrace1(RECVP, TR_ERRORS,
                "auth_server_init: Could not open enable llp #%d",i);
        Kabort("Exiting...");
      }
    }
#endif
}

static xkern_return_t 
recvauthOpenDone( self,lls, llp,hlpType)
     XObj self,lls, llp,hlpType;
{
  PSTATE *ps = (PSTATE *)self->state;
  char buf[256];
  char idbuf[100];
  int  idlen, i;
  Part p[2];
  Enable *e;
  XObj s;
  SSTATE *ss;
  xkern_return_t val;


  if (xControl(lls, GETPARTICIPANTS, buf, 256) < 1) {
    xTrace0(RECVP, TR_ERRORS, "recvauthOpenDone - getparticpants fails");
    return XK_FAILURE;
  }
  
  partInternalize(p, buf);
  idlen = part2ToId(p, idbuf, 100);

  if ( (idlen != ps->paslen) || 
       (mapResolve(ps->pasMap, idbuf, &e) != XK_SUCCESS) ) {
    xTrace0(RECVP, TR_ERRORS, "recvauthOpenDone -- no passive open");
    return XK_FAILURE;
  }

  xDuplicate(lls);

  idlen = partToId(p, idbuf, 100);
  if (ps->taglen == 0) {
    /* first opendone */
    ps->taglen = idlen;
    ps->tagMap = mapCreate(MAP_SZ, idlen);
  }

  if (ps->taglen != idlen) {
    xTrace0(RECVP,TR_ERRORS,
	    "recvauthOpenDone -- idlen differs from ps->taglen");
    return XK_FAILURE;
  }

  if (mapResolve(ps->tagMap, idbuf, &s) != XK_SUCCESS) {
    /* first opendone for this connection */
    s = xCreateSessn(recvauth_sessn_init, e->hlpRcv, e->hlpType, self, 0,0);
    ss = (SSTATE *)xMalloc(sizeof(SSTATE));
    if (!ss) {
      xTrace0(RECVP, TR_ERRORS,
	      "recvauthOpenDone -- malloc failed for SSTATE");
      return XK_FAILURE;
    }
    bzero((char*)ss, sizeof(SSTATE));
    s->state = (void *)ss;
    ss->lower[0] = lls;
    ss->curr = 1;
    ss->tag_bind = mapBind(ps->tagMap, (char *)idbuf, (int)s);
    if (xControl(lls, JOINSETCONTROL, (char*)&s, sizeof(XObj)) !=
        sizeof(XObj)) {
      xTrace0(RECVP, TR_ERRORS, "could not set control session");
      return XK_FAILURE;
    }
    return XK_SUCCESS;
  }

  ss = (SSTATE *)s->state;
  ss->lower[ss->curr++] = lls;

  if (ss->curr == 4) {
    /* received all the opendones */
    mapRemoveBinding(ps->tagMap, ss->tag_bind);
    for(i=0;i<4;i++) {
      mapBind(ps->activeMap, (char*)&ss->lower[i], (int)s);
    }
#if 0
    val = xOpenDone(e->hlpRcv, s, self);
    if (val != XK_SUCCESS) {
      xTrace0(RECVP, TR_ERRORS, "recvauthOpenDone -- could not opendone hlp");
      for(i=0;i<4;i++) {
	xClose(ss->lower[i]);
      }
      return XK_FAILURE;
    }
#endif
  }

  return XK_SUCCESS;

}

static xkern_return_t recvauthOpenEnable( self, hlpRcv, hlpType, part)
     XObj self, hlpRcv, hlpType;
     Part *part;
{
  char buf[256];
  char idbuf[100];
  int  idlen, i, retval;
  PSTATE *ps = (PSTATE *)self->state;
  Enable *e;
  int buf_len = 256;

  /* need to save a copy of the part list */
  if (partExternalize(part, buf, &buf_len) != XK_SUCCESS) {
    xTrace0(RECVP, TR_ERRORS,
	    "recvauthOpenEnable -- partExternalize failed!");
    return XK_FAILURE;
  }

  idlen = partToId(part, idbuf, 100);


  if (ps->paslen == 0) {
    /* this is the first openenable */
    ps->paslen = idlen;
    ps->pasMap = mapCreate(MAP_SZ, idlen);
  }

  if(ps->paslen != idlen) {
    xTrace0(RECVP, TR_ERRORS, "recvauthOpenEnable - bad idLen");
    return XK_FAILURE;
  } 
  
  if (mapResolve(ps->pasMap, idbuf, &e) != XK_FAILURE) {
    xTrace0(RECVP, TR_ERRORS,
	    "recvauthOpenEnable -- duplicate openenable");
    return XK_FAILURE;
  }

  for (i=0; i < 4; i++) {
    partInternalize(part, buf);
    retval = xOpenEnable(self, hlpType, xGetDown(self, i), part);
    if (retval != XK_SUCCESS) {
      xTrace1(RECVP, TR_ERRORS,
	      "recvauthOpenEnable -- could not openenable llp #%d", i);
      return XK_FAILURE;
    }
  }

  e = (Enable *)xMalloc(sizeof(Enable));
  if (!e) {
    xTrace0(RECVP, TR_ERRORS,
	    "recvauthOpenEnable -- malloc failed for enable object");
    return XK_FAILURE;
  }
  e->hlpRcv = hlpRcv;
  e->hlpType = hlpType;
  e->binding = mapBind(ps->pasMap, (char *)idbuf, (int) e);

  if (e->binding == ERR_BIND) {
    xTrace0(RECVP, TR_ERRORS,
	    "recvauthOpenEnable -- could not bind enable object");
    xFree((void *)e);
    return XK_FAILURE;
  }

  xTrace0(RECVP, TR_ERRORS,"recvauthOpenEnable -- exiting");
  return XK_SUCCESS;
}

static xkern_return_t recvauthDemux( self, lls, msg )
     XObj self, lls;
     Msg *msg;
{
  PSTATE *ps = (PSTATE *)self->state;
  int len = msgLen(msg);
  XObj sessn;
  int val;

  xTrace1(RECVP, TR_MAJOR_EVENTS, "recvauthDemux entered -- message size %d",
	  len);

  if (mapResolve(ps->activeMap, &lls, &sessn) != XK_SUCCESS) {
    xTrace0(RECVP, TR_ERRORS,"recvauthDemux -- lls not found in active map");
    return XK_FAILURE;
  }

  val = xPop(sessn, lls, msg, 0);

  xTrace0(RECVP, TR_MAJOR_EVENTS, "recvauthDemux exiting");

  return val;
}

static xkern_return_t recvauthPop(s, lls, msg ,h)
     XObj s, lls;
     Msg *msg;
     void *h;
{
  SSTATE *ss = (SSTATE *)s->state;
  int i, j, num, retval;
  krb5_data packet;
  krb5_enc_tkt_part *enc_tkt_part;
  krb5_keyblock *keyblock;
  xkern_return_t val;

  j = -1;
  for (i=0; i<4; i++) {
    if (ss->lower[i] == lls) {
      j = i;
      break;
    }
  }

  if (j == -1) {
    /* something has gone really bad */
    xTrace0(RECVP, TR_ERRORS, "recvauthPop -- lls not found");
    return XK_FAILURE;
  }

  ss->m[j] = msg;
  
  if (j == 2) {
    /* this should be the decrypted ticket */
    /* need to place the key in the km */
    packet.length = msgLen(msg);
    packet.data = (char *)xMalloc(packet.length);
    if ( !packet.data ) {
      xTrace0(RECVP, TR_ERRORS, "malloc failed for packet.data");
      return XK_FAILURE;
    }
    
    msgPop(msg, mycopypop, packet.data, packet.length, 0);
    
    /* now I need to decode the packet */
    retval = decode_krb5_enc_tkt_part(&packet, &enc_tkt_part);
    if (retval) {
      xTrace1(RECVP, TR_ERRORS,
	      "error(%d) occurred while decoding enc_tkt_part", retval);
      return retval;
    }
    ss->enc_tkt_part = enc_tkt_part;
    keyblock = enc_tkt_part->session;
    retval = xControl(ss->lower[3], CRYPT_IN_SET,
		      (char*)keyblock->contents,keyblock->length);
    if (retval == -1) {
      xTrace0(RECVP, TR_ERRORS, "CRYPT_IN_SET failed for session key");
      return -1;
    }
    
    retval = xControl(ss->lower[3], CRYPT_OUT_SET,
		      (char*)keyblock->contents,keyblock->length);
    if (retval == -1) {
      xTrace0(RECVP, TR_ERRORS, "CRYPT_OUT_SET failed for session key");
      return -1;
    }
    return XK_SUCCESS;
  } else if (j == 3) {
    /* add code here */
    /* need to check the authenticator against the ticket */
    
    val = xOpenDone(xGetUp(s), s, s->myprotl);
    if (val != XK_SUCCESS) {
      xTrace0(RECVP, TR_ERRORS,"recvauthOpenDone -- could not opendone hlp");
      for(i=0;i<4;i++) {
	xClose(ss->lower[i]);
      }
      return XK_FAILURE;
    }
  }
  return XK_SUCCESS;
}

static void           
recvauth_sessn_init( self )
     XObj self;
{
  self->pop = recvauthPop;
  self->control = recvauthSessnControl;
}

static int            
recvauthProtlControl( self, op, buf, len )
     XObj self;
     int op, len;
     char *buf;
{
  return xControl(xGetDown(self,0), op,buf,len);
}

static int            
recvauthSessnControl( self, op, buf, len )
     XObj self;
     int op, len;
     char *buf;
{
  SSTATE *ss = (SSTATE *)self->state;
  int real_len;

  if (op == JOINDONE) {
    return process_ap_request(self);
  } else if (op == OBTAINKEY) {
    if (ss->enc_tkt_part) {
      real_len = ss->enc_tkt_part->session->length;
      if (len >= real_len) {
	bcopy(ss->enc_tkt_part->session->contents, buf,real_len);
	return real_len;
      } else {
	return -1;
      }
    }
  }
  return xControl(ss->lower[0], op,buf,len);
}

static void mycopypush(src, dst, count, arg)
     void *src, *arg;
     char *dst;
     long  count;
{
  bcopy((char *)src, dst, count);
}

static long mycopypop(dest, src, count, arg)
     void *dest;
     char *src;
     long count;
     void *arg;
{
  bcopy(src, (char*)dest, count);
  return count;
}


static int process_ap_request(self)
     XObj self;
{
  SSTATE *ss = (SSTATE *)self->state;
  krb5_ap_req *ap_req;
  krb5_data    scratch, scratch2;
  krb5_authenticator *authent;
  int retval;
  Msg rep_msg;
  


  msgPop(ss->m[1], mycopypop, (char *)&ap_req, sizeof(krb5_ap_req *), 0);

  scratch.length = msgLen(ss->m[3]);
  scratch.data = xMalloc(scratch.length);
  if ( !scratch.data ) {
    xTrace0(RECVP, TR_ERRORS,"malloc failed for scratch");
    return XK_FAILURE;
  }

  msgPop(ss->m[3], mycopypop, (char *)scratch.data, scratch.length, 0);

  retval = decode_krb5_authenticator(&scratch, &authent);
  if (retval) {
    xTrace1(RECVP, TR_ERRORS,
	    "error (%d) received while decoding authenticator", retval);
    return XK_FAILURE;
  }

  if (!krb5_principal_compare(ss->enc_tkt_part->client, authent->client)) {
    xTrace0(RECVP, TR_ERRORS,
	    "client portions in ticket and authenticator don't match");
    return XK_FAILURE;
  }
  /* check to see if everything worked out ok */

  /* if everything is okay, return a message with length zero */


  /* how to construct ok message */
  scratch2.length = 0;
  msgConstructEmpty(&rep_msg);
  msgPush(&rep_msg, mycopypush, (void*)&scratch2.length, sizeof(int), 0);

  /* tell kerb_serv_bottom to change status */
  xControl(ss->lower[0], END_AUTHENTICATION, 0,0);

  xPush(ss->lower[1], &rep_msg);
  xControl(ss->lower[0], JOINDONE, 0,0);
  return XK_SUCCESS;
}


static int
partToId(p, buf, maxLen)
    Part *p;
    char *buf;
    int maxLen;
{
    int i,j,len;
    char *ptr;
    int numPart = partLen(p);
    u_char *ptr2;

    xTrace0(RECVP, TR_EVENTS, "partToLen - entered");
    len = 0;
    ptr = (char *)buf;
    for (j=0;j<numPart; j++) {
        for (i=p[j].stack.top; i>0; i--)  {
                len += p[j].stack.arr[i-1].len;
                if (len > maxLen) return 0;
                bcopy((char *)p[j].stack.arr[i-1].ptr,(char *)ptr, p[j].stack.arr[i-1].len);
                ptr += p[j].stack.arr[i-1].len;
        }
    }

    xTrace1(RECVP, TR_EVENTS, "partToLen - exiting len %d",len);
    xIfTrace(RECVP, TR_EVENTS) {
        printf("part2ToLen - id= (");
        ptr2 = (u_char *)buf;
        for (i=0;i<len;i++) printf("%x",ptr2[i]);
        printf(")\n");
    }

    return(len);
}




static int
part2ToId(p, buf, maxLen)
    Part *p;
    char *buf;
    int maxLen;
{
    int i,len;
    char *ptr;
    u_char *ptr2;
    
    xTrace0(RECVP, TR_EVENTS, "part2ToLen - entered");
    len = 0;
    ptr = (char *)buf;
    for (i=p[0].stack.top; i>0; i--)  {
       	len += p[1].stack.arr[i-1].len;
	if (len > maxLen) return 0;
       	bcopy((char *)p[1].stack.arr[i-1].ptr,(char *)ptr, p[1].stack.arr[i-1].len);
       	ptr += p[1].stack.arr[i-1].len;
    }
    xTrace1(RECVP, TR_EVENTS, "part2ToLen - exiting len %d",len);
    xIfTrace(RECVP, TR_EVENTS) {
        printf("part2ToLen - id= (");
        ptr2 = (u_char *)buf;
        for (i=0;i<len;i++) printf("%x",ptr2[i]);
        printf(")\n");
    }

    return(len);
}
