/*
 * $RCSfile: ipf.c,v $
 *
 * To use this protocol define IPF_PROT and include this source file.
 * 
 * x-kernel v3.2
 *
 * Copyright (c) 1996  Arizona Board of Regents
 *
 * $Log: ipf.c,v $
 * Revision 1.1  1997/04/24 20:06:04  mjk
 * Initial revision
 *
 */

#include "xkernel.h"
#include "xk_debug.h"
#include "x_stdio.h"
#include "netmask.h"
#include "eth_i.h"
#include "ip.h"



/* 
 * Defines 
 */

				/* standard quoting macros 		*/
#ifdef XQUOTE
#undef XQUOTE
#endif
#ifdef QUOTE
#undex QUOTE
#endif
#ifdef XPASTE
#undef XPASTE
#endif
#ifdef PASTE
#undef PASTE
#endif

#define XQUOTE(a)	#a
#define QUOTE(a)	XQUOTE(a)

#define XPASTE(a,b)	a##b
#define PASTE(a,b)	XPASTE(a,b)

#define IPF_NAME	PASTE(ipf_,IPF_PROT)
#define IPF_QNAME	QUOTE(IPF_NAME)


#define IPF_BUFSIZE	1024	/* used for xControl ops 		*/

#ifdef XK_DEBUG
/* #define XK_RX_DEBUG */
/* #define XK_TX_DEBUG */
#endif


/* 
 * Typedefs 
 */

typedef struct {		/* Protocol State 			*/
  IPhost	ip_server;		/* real IP 			*/
  IPhost	ip_client;		/* bogus IP 			*/
} PState;


/* 
 * Prototypes 
 */

void			PASTE(IPF_NAME,_init)(XObj);
static xkern_return_t	PASTE(IPF_NAME,_demux)(XObj, XObj, Msg *);
static xkern_return_t   PASTE(IPF_NAME,_rom_server)(XObj, char **,
						    int, int, void *);
static xkern_return_t   PASTE(IPF_NAME,_rom_client)(XObj, char **,
						    int, int, void *);
#ifdef XK_DEBUG
static void		PASTE(IPF_NAME,_dump)(XObj, FILE *, const char *,
					      unsigned char *, int);
#endif

			      


/* 
 * Globals 
 */

int PASTE(trace,PASTE(IPF_NAME,p));

					/* rom options			*/
static XObjRomOpt	PASTE(IPF_NAME,_opts)[] = {
  { "server",	3, PASTE(IPF_NAME,_rom_server) },
  { "client",	3, PASTE(IPF_NAME,_rom_client) }
};



/* 
 * Public Functions
 */

void
PASTE(IPF_NAME,_init)(XObj self)
{
  XObj		llp;
  Part		part;
  PState	*ps;

  xTraceP0(self, TR_MAJOR_EVENTS, "init");

  if ( !xIsProtocol(llp = xGetDown(self, 0)) ) {
    xError(IPF_QNAME " cannot get lower protocol");
    return;
  }

  ps = X_NEW(PState);
  self->state = ps;

				/* Parse ROM file 			*/
  findXObjRomOpts(self, PASTE(IPF_NAME,_opts),
		  (sizeof PASTE(IPF_NAME,_opts)) / sizeof(XObjRomOpt), NULL);  
 

  self->demux = PASTE(IPF_NAME,_demux);

  partInit(&part, 1);
  partPush(part, ANY_HOST, 0);

  if ( xOpenEnable(self, self, llp, &part) == XK_FAILURE ) {
    xError(IPF_QNAME " cannot openenable lower protocol");
  }

} /* PASTE(IPF_NAME,_init) */



/* 
 * Public Protocol Methods 
 */

static xkern_return_t
PASTE(IPF_NAME,_demux)(XObj self, XObj lls, Msg *packet)
{
  PState	*ps		= (PState *)self->state;
  char		buf[IPF_BUFSIZE];
  Part		part[2];
  Sessn 	session;
  IPhost	ip_source;
  IPhost	ip_dest;
  IPhost	ip_nsource;
  IPhost	ip_ndest;
#ifdef XK_DEBUG  
  static char		buf2[0x4000];
#endif

  
  xTraceP0(self, TR_MAJOR_EVENTS, "demux");

#ifdef XK_RX_DEBUG
  xIfTraceP(self, TR_MAJOR_EVENTS) {
    msg2buf(packet, buf2);
    PASTE(IPF_NAME,_dump)(self, stdout, "RX", buf2, msgLen(packet));
  }
#endif  

  xControl(lls, IP_GETPSEUDOHDR, buf, sizeof buf);
  ip_source = ((IPpseudoHdr *)buf)->dst;
  ip_dest   = ((IPpseudoHdr *)buf)->src;

  xTraceP1(self, TR_MAJOR_EVENTS, "source:     %s", ipHostStr(&ip_source));
  xTraceP1(self, TR_MAJOR_EVENTS, "dest:       %s", ipHostStr(&ip_dest));

  /* If the packet is from the UNIX side (client) change the source
     address to that of the xkernel (server) */
  if ( IP_EQUAL(ip_source, ps->ip_client) ||
       IP_EQUAL(ip_source, ps->ip_server) )
  {
    ip_nsource = ps->ip_server;
    ip_ndest   = ip_dest;
  }

  /* If the packet is for the xkernel (server) change the destination
     address to that of the UNIX side (client).  Broadcast packets on
     the xkernel network will also have the destination address
     changed to prevent the xkernel from rebroadcasting onto the
     network.  */
  else if ( IP_EQUAL(ip_dest, ps->ip_server) ||
	    netMaskIsBroadcast(&ip_dest) )
  {
    ip_nsource = ip_source;
    ip_ndest   = ps->ip_client;
  }

  /* Otherwise we should not have received this packet so drop it.*/
  else {
    xTraceP0(self, TR_MAJOR_EVENTS, "dropped");
    return XK_SUCCESS;
  }

  xTraceP1(self, TR_MAJOR_EVENTS, "new source: %s", ipHostStr(&ip_nsource));
  xTraceP1(self, TR_MAJOR_EVENTS, "new dest:   %s", ipHostStr(&ip_ndest));

  partInit(part, 2);		/* new participants 			*/
  partPush(part[0], &ip_ndest, sizeof(IPhost));		/* remote 	*/
  partPush(part[1], &ip_nsource, sizeof(IPhost));	/* local	*/

#ifdef XK_TX_DEBUG
  xIfTraceP(self, TR_EVENTS) {
    msg2buf(packet, buf2);
    PASTE(IPF_NAME,_dump)(self, stdout, "TX", buf2, msgLen(packet));
  }
#endif  
  
  session = xOpen(self, self, xGetDown(self, 0), part);
  if ( session != ERR_XOBJ ) {
    if ( xPush(session, packet) == XK_FAILURE )
      return XK_FAILURE;
    if ( xClose(session) == XK_FAILURE)
      return XK_FAILURE;
  }
  else {
    xTraceP0(self, TR_ERRORS, "cannot open IP session");
    return XK_FAILURE;
  }
  
  return XK_SUCCESS;
} /* PASTE(IPF_NAME,_demux) */


/* 
 * Private Protocol methods 
 */
#ifdef XK_DEBUG
static void
PASTE(IPF_NAME,_dump)(XObj self, FILE *fout, const char *label,
		      unsigned char *pkt, int len)
{
  PState	*ps = (PState *)self->state;
  int		i, j;

  xAssert(fout);
  xAssert(pkt);
  xAssert(label);
  xAssert(len >= 0);

	  /* Dumping is in the debug hex/ascii style and is setup for
	     80 columns.  This puts 16 bytes per line. */
 
  for (i=0; i<len; i+=16) {
    char	buffer[80];
    char	*p = &buffer[0];
    char	*q = &buffer[52];
    
    memset(buffer, (int)' ', sizeof buffer);
    
    for (j=0; j<16 && (j+i) < len; j++) {
      unsigned char	byte = pkt[i+j];

      if ( j == 8 )
	p += sprintf(p, "- ");
      
      p += sprintf(p, "%02x ", byte);
      
      if ( isprint(byte) )
	*q++ = byte;
      else
	*q++ = '.';
    }

    *p = ' ';
    *q = '\0';
    xTraceP2(self, TR_GROSS_EVENTS, "%s: %s", label, buffer);
  }
} /* PASTE(IPF_NAME,_dump) */
#endif



/* 
 * Private ROM file methods. 
 */

static xkern_return_t
PASTE(IPF_NAME,_rom_server)(XObj self, char **argv,  int argc, int line,
			    void *arg)
{
  PState	 *ps = (PState *)self->state;

  if ( argc == 3 && str2ipHost(&ps->ip_server, argv[2]) == XK_SUCCESS ) {
    xTraceP1(self, TR_MAJOR_EVENTS, "server is %s", ipHostStr(&ps->ip_server));
    return XK_SUCCESS;
  }
  return XK_FAILURE;
} /* PASTE(IPF_NAME,_rom_server) */


static xkern_return_t
PASTE(IPF_NAME,_rom_client)(XObj self, char **argv,  int argc, int line,
			    void *arg)
{
  PState	 *ps = (PState *)self->state;

  if ( argc == 3 && str2ipHost(&ps->ip_client, argv[2]) == XK_SUCCESS ) {
    xTraceP1(self, TR_MAJOR_EVENTS, "client is %s", ipHostStr(&ps->ip_client));
    return XK_SUCCESS;
  }
  return XK_FAILURE;
} /* PASTE(IPF_NAME,_rom_client) */


