/* 
 * serial.c
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993 Bryce Nutter
 *
 *
 */

#include "serial.h"

XObj   The_sess,The_type;;
extern void dumb_func();

void
init_serial_blocks()
{
    int 	i;
    block 	*bp;
    Msg 	*msg;
    static int	initialized = 0;
    
    if ( initialized ) {
	xTrace0(serialp, TR_EVENTS, "serialblocks already initialized");
	return;
    }

    xTrace0(serialp, TR_EVENTS, "SERIAL - Entering init_serial_blocks");
    initialized = 1;
    pool.total_blocks = MAX_SERIAL_BLOCKS;
    pool.next_block = 0;
    
    bp = &(pool.blocks[0]);
    
    for (i=0; i<MAX_SERIAL_BLOCKS; i++) {
  
      if (!(bp->data_buf = (u_char *) xMalloc(2*(DEVICE_MTU+TLEN+HLEN))))
	Kabort("Malloc of serial packets-init_serial_blocks,A"); 
      clear_buf((u_char *)bp->data_buf,2*(DEVICE_MTU+TLEN+HLEN));

	if (!(msg = (Msg *)xMalloc(sizeof(Msg))))
          Kabort("Malloc of serial packets-init_serial_blocks,B"); 

	msgConstructAllocate(msg, 2*(DEVICE_MTU+TLEN+HLEN), (char **)&(bp->data));
	bp->msg = msg;
	bp->ref = 0;
	bp->id = i;
	bp->cur_len = 0;
	semInit(&bp->sem, 0);
	
	evDetach( evSchedule(block_handler, bp, 0));
        if (i != (MAX_SERIAL_BLOCKS - 1))
	  bp = &(pool.blocks[i + 1]);
    }
    xTrace0(serialp, TR_EVENTS, "SERIAL - Exiting init_serial_blocks");
}



static void
block_handler(ev, arg)
    Event	ev;
    VOID 	*arg;
{
  Msg      *msg;
  XObj     up_obj,sself,up_type,s;
  ActiveId activeid;

  block *bp = (block *)arg;

     for (;;) {
	 semWait(&bp->sem);		/* wait for incoming data */
	 xAssert(bp->cur_len > 0 && bp->cur_len <= (DEVICE_MTU+TLEN+HLEN));
	 xTrace1(serialp, TR_EVENTS, "SERIAL - thread for block %x runs",
		 (int)bp);
	 sself  = bp->self;
	 up_obj = The_sess;
	 up_type = The_type;
	 msg = bp->msg;

	 s = serialCreateSessn(sself, The_sess, &activeid);

	 msgConstructInplace(msg,(char *)bp->data,bp->cur_len,dumb_func);
	 xDemux(s,msg);

	 xTrace1(serialp, TR_EVENTS, "SERIAL - thread for block %x returns from xDemux",(int)bp);

	 if ( --activeThreads > activeHighWater ) {
	     activeHighWater = activeThreads;
	 }

	 msgDestroy(bp->msg);

	 bp->cur_len = 0;

	 msgConstructAllocate(bp->msg, 2*(DEVICE_MTU+TLEN+HLEN), (char **)&(bp->data));
       
	 clear_buf((u_char *)bp->data_buf,2*(DEVICE_MTU+TLEN+HLEN));

	 bp->ref = NOREF; 
     }
}




/*
 * PPP_handle_msg--buflen = PPP_handle_msg(TTYB, bp->data)
*/
int PPP_handle_msg(ttytype,data,data_buf)
     int ttytype;
     u_char *data;
     u_char *data_buf;
{
  int ct=0,done=0;
  u16 Crc,Prev_Crc;
  long len;
  u_char *ptr,*track;
  int        now,k,r_b_ptr,alter_next=0;
  
  ptr = data_buf;
  track = ptr;
  hand++;
  ct = 0;
  ptr = data;

  xTrace0(serialp, TR_FULL_TRACE, "SERIAL - Entering PPP_handle_msg");

/*  printf("%c",*data_buf);*/
  if(*data_buf++ == FLAG) {
      now = 1;
      alter_next = 0;
      r_b_ptr = 1;
      *data++ = FLAG;
      ct = 0;
      done = 0;
      do {
	  ct++;
	  /*printf("%c",*data_buf);*/
	  now++;
	  if(alter_next) {
	      alter_next = 0;
	      *data++ = (*data_buf ^ XOR);
	      data_buf++;
	      r_b_ptr++;
	    }
	  else {
	      if ( *data_buf != FLAG) {
		  if (*data_buf == CEO) {
		      alter_next = 1;
		      data_buf++;
		    }
		  else {
		      *data++ = *data_buf++;
		      r_b_ptr++;
		    }
		}
	      else {
		  *data++ = *data_buf++;
		  r_b_ptr++;
		  done = 1;
		}

	    }
	} while( !done ); 
      /*printf("\n");*/

      len = r_b_ptr;
      Crc = PPPINITFCS;
      k = len-4;
      Crc = pppfcs(Crc,&ptr[1],k);

      bcopy((char*)&ptr[len-3],(char*)&Prev_Crc,2);

      if(Crc != Prev_Crc) {
	xTrace0(serialp, TR_MAJOR_EVENTS, "CRC FAILURE");
	printf("PPP - CRC FAILURE\n");
	len = -1;
      }

      k = len;
      leng = len;
    }
  else {
      printf("FIRST FLAG NOT FOUND!\n");
      return -1;
    }
  reset_again = 0;

  xTrace0(serialp, TR_FULL_TRACE, "SERIAL - Exiting PPP_handle_msg");
  return len;
}


static block *
nextBlock()
{
    block 	*bp;
    int		i;

    xTrace0(serialp, TR_EVENTS, "SERIAL - getting next block");
    i = pool.next_block;
    do {
	bp = &(pool.blocks[i]);
	i++;
	i %= pool.total_blocks;
    } while ( bp->ref == 1 && i != pool.next_block );
    if ( bp->ref == INUSE ) {
        xTrace0(serialp, TR_EVENTS, "SERIAL - all threads busy");
	return NULL;
    }

    bp->ref = INUSE;
    pool.next_block = i;
    return bp;
}



/* 
 *
 */
static int
readserial2demux( arg )
    VOID	*arg;
{
    int		buflen = 0,k,rval,old_amount;
    int         left_over;
    int         Complete,Found,Chop,count,new_amount,trash,let_it_go;
    block 	*bp,*let_bp;
    u_char      *ptr;
    
    xTrace0(serialp, TR_EVENTS, "SERIAL - readserial2demux");
    if(!Done) {
        xTrace0(serialp, TR_EVENTS, "SERIAL - Readserial2demux:NOT DONE");
	bp = old_bp;
	old_amount = bp->cur_len;
      }
    else 
      if ( ! (bp = nextBlock()) ) { 
	  xError("SERIAL ERROR: Can't get next buffer,dropping incoming packet");
	  return -1;
	}
      else
	old_amount = 0;

    ptr = bp->data_buf;
    if(!Done)
      ptr += (bp->cur_len);

    rval = read(tty_fd,ptr,DEVICE_MTU+TLEN+HLEN);
/*    printf("just read %d bytes\n",rval);

    trk = ptr;
    for (i=0;i<rval;i++)
      {
	printf("%c",*trk);
	trk++;
      }
    printf("\n");*/

    ptr = bp->data_buf;

    new_amount = old_amount + rval;
    Complete = 0;
    count = 0;

    do {
	Found = 0;
	count = 0;
	if(*ptr != FLAG)    
	    do {
		count++;
		ptr++;
	      } while ((*ptr != FLAG) && (count != new_amount));
	if (count > 0)
	    Chop = count;
	else
	    Chop = 0;
      
	if (count == new_amount) {
	    Complete = 1;
	    Done = 1;
	  }
	else {
	    let_it_go = 0;
	    do {
		ptr++;
		count++;
	      } while((count < new_amount) && (*ptr != FLAG));

	    if (*ptr == FLAG) {
		Found = 1;
		(bp->data_buf) += Chop;
		let_it_go = 1;
		let_bp = bp;
	      }

	    if((count == new_amount) && (Found)) {
		Complete = 1;
		Done = 1;
	      }
	    else
	      if(!Found) {
		  left_over = new_amount;
		  old_bp = bp;
		  bp->cur_len = left_over;
		  Complete = 1;
		  Done = 0;
		}
	      else {
		  ptr++;
		  left_over = new_amount - (count + 1);
		  trash = 0;
		  if(left_over != 0) {
		      if (*ptr != FLAG) {
			  trash = 0;
			  do {
			      trash++;
			      ptr++;
			    } while ((*ptr != FLAG) && (trash != left_over));
			}
		    }
		  else {
			left_over = 0;
			trash = 0;
		      }
		  if (trash == left_over) {
		      Complete = 1;
		      Done = 1;
		    }
		  else {
		      left_over -= trash;
		      if ( ! (bp = nextBlock()) ) {
			  xError("SERIAL ERROR: Can't get next buffer,dropping packet");
			  return -1;
			}
		      bcopy((char *)ptr,(char *)bp->data_buf,left_over);
		      new_amount = left_over;
		      ptr = bp->data_buf;
		    }
		}
	    if(let_it_go) {
		buflen = PPP_handle_msg(TTYB, let_bp->data,let_bp->data_buf);
		if(buflen != -1) {
		    let_bp->cur_len = buflen;
		    let_bp->self = (XObj)arg;
		    if ( ++activeThreads > activeHighWater ) {
		      activeHighWater = activeThreads;
		    }
		    k = buflen;
		    xTrace1(serialp, TR_SOFT_ERRORS, "SERIAL - readserial2demux - just read %d bytes from device.",new_amount);
		    xTrace0(serialp, TR_EVENTS, "SERIAL - readserial2demux - signalling a shepard ");
		    semSignal(&let_bp->sem);
		  }
		else {
		  	 let_bp->cur_len = 0;
			 clear_buf((u_char *)let_bp->data_buf,2*(DEVICE_MTU+TLEN+HLEN));
			 let_bp->ref = NOREF;
		}
		let_it_go = 0;
	      }
	  }
      } while (!Complete); 
 
    xTrace0(serialp, TR_EVENTS, "SERIAL - Exiting readserial2demux");
    return 0;
  }



void init_send()
{
  int res;
  xTrace0(serialp, TR_EVENTS, "Serial - Entering init_send");
  if (!already_opened) {
      already_opened = 1;
      tty_fd = open(DEVICEA,O_RDWR);

      if (ioctl(tty_fd,TCGETS,&tio) < 0)
	perror("IOCTL,TCGETS");
      tio.c_lflag &= ~(ICANON|ISIG|ECHO);
      tio.c_cc[VTIME] = 0;
      tio.c_cc[VMIN] = CHUNK;
      
      tio.c_cflag  = CRTSCTS | B38400;
      tio.c_cflag |= CS8|CREAD|HUPCL;
      tio.c_iflag = IGNBRK;
      tio.c_oflag &= ~OPOST;

      if (ioctl(tty_fd,TCSETS,&tio) < 0)
	perror("IOCTL,TCSETS");

      if ((res = fcntl(tty_fd,F_SETOWN,getpid())) < 0)
	printf("fcntl(F_SETOWN) returned:%d\n",res);
      if ((res = fcntl(tty_fd,F_SETFL,FASYNC)) < 0)
	perror("fcntl(F_SETFL), FASYNC");
    }
    xTrace0(serialp, TR_EVENTS, "SERIAL - Exiting init_send");
}


void
ppp_serial_init( self )
    XObj	self;
{
    int res;
    xTrace0(serialp, TR_EVENTS, "SERIAL - ppp_serial_init");
    if ( instance == 0 ) {
	init_serial_blocks();		
    }
    if ( instance > PPP_MAX_INSTANCES ) 
	Kabort("SERIAL - too many instances");

    Init_blocks();
    if ((res = fcntl(tty_fd,F_SETOWN,getpid())) < 0)
      xTrace1(serialp, TR_ERRORS, "SERIAL serial_init: fcntl returned %d",
	    res);
    if ((res = fcntl(tty_fd,F_SETFL,FASYNC)) < 0)
      perror("fcntl(F_SETFL), FASYNC");
    instance++;
    xTrace0(serialp, TR_EVENTS, "SERIAL - Exiting ppp_serial_init");
}



/*
 *   serial_init
 */
int
serial_init(self)
    XObj self;
{
    Part	part;
    SERIALPSTATE	*pstate;

    xTrace0(serialp, TR_EVENTS, "SERIAL - Entering serial_init");
    xAssert(xIsProtocol(self));
    getproc_protl(self);

    pstate = X_NEW(SERIALPSTATE);
    self->state = (VOID *)pstate;
    pstate->activemap = mapCreate(ACTIVE_MAP_SIZE, sizeof(ActiveId));
    pstate->passivemap = mapCreate(PASSIVE_MAP_SIZE, sizeof(PassiveId)); 

    partInit(&part, 1);

    read_rom(pstate);
    init_send();
    installSignalHandler(tty_fd, readserial2demux, self);
    ppp_serial_init();
    partPush(part, ANY_HOST,sizeof(Part));
    xTrace0(serialp, TR_EVENTS, "SERIAL - serial_init done");
    return 0;
}

/*
 *   serialOpen
 */
static XObj
serialOpen( self, hlpRcv, hlpType, p )
    XObj self, hlpRcv, hlpType;
    Part *p;
{

    XObj	serial_s;
    xkern_return_t serial_s_ret;
    ActiveId 	key;
    SERIALPSTATE	*pstate;
    int         num_ifs=0;

    xTrace0(serialp, TR_EVENTS, "SERIAL - Entering serialOpen");
    pstate = (SERIALPSTATE *)self->state;
    key.device_name = pstate->interface[num_ifs].device;

    xTrace1(serialp, TR_MAJOR_EVENTS, "SERIAL open: to device %d",
	    key.device_name);
      
    partInit(p, 1);
    partPush(*p, &key.device_name,sizeof(int));

    serial_s_ret = mapResolve(pstate->activemap,(char*)&key,(void **)&serial_s);

    if(serial_s_ret != XK_SUCCESS) 
	serial_s = serialCreateSessn(self,hlpRcv,&key);

    xTrace0(serialp, TR_EVENTS, "SERIAL - Exiting serialOpen");
    return (serial_s);

}
/*
 *   serialCreate Sessn
 */
static XObj
serialCreateSessn(self, hlp, key)
    XObj self;
    XObj hlp;
    ActiveId *key;
{
    XObj	s,s_ret;
    SERIALSSTATE	*sstate;
    SERIALPSTATE	*pstate;

    xTrace0(serialp, TR_EVENTS, "SERIAL - Entering serialCreateSessn");
    pstate = (SERIALPSTATE *)self->state;
    s = xCreateSessn(getproc_sessn, hlp, self,self,0,0);
    xAssert(xIsSession(s));

    s->binding = (Bind)mapBind(pstate->activemap, (char *)key, (int)s);

    if ( s->binding == ERR_BIND ) {
	xTrace0(serialp, TR_EVENTS, "SERIAL create sessn - session already existed");
	serialCloseSessn(s);
	mapResolve(pstate->activemap, (char *)key, (void **)&s_ret);
	return s_ret; 
    }

    if (!(sstate = (SERIALSSTATE *)xMalloc(sizeof(SERIALSSTATE))))
      Kabort("Malloc of serial state-serialCreateSessn"); 
xTrace0(serialp, TR_MAJOR_EVENTS,
		"SERIAL create sessn ");
    s->state = (char *)sstate;
    xTrace0(serialp, TR_EVENTS, "SERIAL - Exiting serialCreateSessn");
    return s;
}

/*
 *   serialControlSessn
 */
int
serialControlSessn(s, opcode, buf, len)
    XObj s;
    int opcode;
    char *buf;
    int len;
{

  SERIALSSTATE        *sstate;
  SERIALPSTATE        *pstate;
  int c;

  xTrace0(serialp, TR_EVENTS, "SERIAL - Entering serialControlSessn");
  
  xAssert(xIsSession(s));
  
  sstate = (SERIALSSTATE *) s->state;
  pstate = (SERIALPSTATE *) s->myprotl->state;
  
  switch (opcode) {
    
  case GETMYPROTO:
    checkLen(len, sizeof(long));
    return sizeof(long);

  case GETPEERPROTO:
    checkLen(len, sizeof(long));
    return sizeof(long);

   case GETPARTICIPANTS:
     {
	 Part	*p = (Part *)buf;
	 int 	retLen;
	 long	*key;

	 retLen = xControl(xGetDown(s, 0), opcode, buf, len);
	 if ( retLen >= sizeof(Part) ) {
	     key = X_NEW(long);
	     partPush(p[0], key, sizeof(Part));
	 }
	 if ( retLen >= 2 * sizeof(Part) ) {
	     key = X_NEW(long);
	     partPush(p[0], key, sizeof(Part));
	 }
	 return retLen;
     }

  case GETMAXPACKET:
    c = (pstate->interface[0].mtu - TLEN - HLEN);
    *(int*)buf = c;
    return c;
  case GETOPTPACKET:
    {
	checkLen(len, sizeof(int));
	if ( xControl(xGetDown(s, 0), opcode, buf, len) < sizeof(int) ) {
	    return -1;
	}
	 *(int *)buf -= (HLEN + TLEN);
	return sizeof(int);
    }
    
  default:
    return xControl(xGetDown(s, 0), opcode, buf, len);
  }
}

/*
 *   serialControlProtl
 */
int
serialControlProtl(self, opcode, buf, len)
    XObj self;
    int opcode;
    char *buf;
    int len;
{

  SERIALPSTATE	*ps = (SERIALPSTATE *)self->state;
  int 	reply,num_ifs=0;
  ArpBinding	*b = (ArpBinding *)buf;
 
  xTrace0(serialp, TR_EVENTS, "SERIAL - Entering serialControlProtl");
  xAssert(xIsProtocol(self));
    
    switch (opcode) {
	
    case GETMAXPACKET:
      reply = (ps->interface[num_ifs].mtu - TLEN - HLEN);
      break;

    case RESOLVE:
      checkLen(len, sizeof(ArpBinding));
      if ( !strcmp((char *)ipHostStr((IPhost*)&b->ip),
		   (char *)ipHostStr((IPhost*)&ps->interface[num_ifs].THEIRPPPAddr_int)) )
	  reply = sizeof(ArpBinding);
      else
	reply = -1;
      break;

    case ARP_GETMYBINDING:
      checkLen(len, sizeof(ArpBinding));
      b->ip = *(IPhost*)&ps->interface[num_ifs].MYPPPAddr_int;
      reply = sizeof(ArpBinding);
      break;
      
    default:
      reply = -1;
	
    }
  xTrace0(serialp, TR_EVENTS, "SERIAL - Exiting serialControlProtl");
  return(reply);

}


/*
 *   serialDemux
 */
static xkern_return_t
serialDemux(bp)
    block *bp;
{

return XK_SUCCESS; 

}


/*
 * serialOpenEnable
 */
static xkern_return_t
serialOpenEnable( self, hlpRcv, hlpType, p )
    XObj self, hlpRcv, hlpType;
    Part *p;
{
    SERIALPSTATE 	*pstate;
    Enable 	*e;
    xkern_return_t e_ret;
    PassiveId   key2;
    int         num_ifs=0;

    xTrace0(serialp, TR_EVENTS, "SERIAL - Entering serialOpenEnable");
    
    pstate = (SERIALPSTATE *)self->state;
    
    The_sess = hlpRcv;
    The_type = hlpType;
    key2.device_name = pstate->interface[num_ifs].device;

    e_ret = mapResolve(pstate->passivemap, (char *)&key2, (void **)&e);
    if ( e_ret == XK_SUCCESS) {
	if ( e->hlpRcv == hlpRcv ) {
	    e->rcnt++;
	    return XK_SUCCESS;
	}
	return XK_FAILURE;
    }
    if (!(e = (Enable *)xMalloc(sizeof(Enable))))
      Kabort("Malloc of serial enable-serialOpenEnable"); 
    e->hlpRcv = hlpRcv;
    e->hlpType = hlpType;
    e->rcnt = 1;
    e->binding = mapBind(pstate->passivemap, (char *)&key2, (int) e);
    if (e->binding == ERR_BIND) {
      xFree((char *)e);
      return XK_FAILURE;
    }
    xTrace0(serialp, TR_EVENTS, "SERIAL - Exiting serialOpenEnable");
    return XK_SUCCESS;
}


static xkern_return_t
serialCloseSessn(s)

    XObj	s;
{
    SERIALPSTATE   *pstate;
    SERIALSSTATE   *sstate;

    xAssert(xIsSession(s));
    xTrace1(serialp, TR_MAJOR_EVENTS, "SERIAL CLOSING session(does zippo) %x", s);
    xAssert(s->rcnt == 0);
    pstate = (SERIALPSTATE *)s->myprotl->state;
    sstate = (SERIALSSTATE *)s->state;
    return XK_SUCCESS;
}


/*
 *   serialCloseProtl
 */
static xkern_return_t
serialCloseProtl(self)
    XObj self;
{
  SERIALPSTATE        *pstate;
  
  xAssert(xIsProtocol(self));
  xTrace0(serialp, TR_MAJOR_EVENTS, "SERIAL - CloseProtl");
  if (((self->rcnt)) <= 0) {
      pstate = (SERIALPSTATE *) self->state;
      mapClose(pstate->activemap);
      mapClose(pstate->passivemap);
      xFree((char *) pstate);
  }
  return XK_SUCCESS;

}

bool
send_data(u_char *msgPtr, long len, void *bufPtr)
{
  int okayed = 0,rval,rval_total=0;
  int Done=0,Sent=0;
  int k;
  u_char  *track;
  fd_set   writefd;
  xTrace0(serialp, TR_EVENTS, "SERIAL - Sending data to serial line");
  track = msgPtr;

    do {
       okayed = 0;
      do {
	FD_ZERO(&writefd);
	FD_SET(tty_fd,&writefd);

	if ((k = select(128,(fd_set*)0, &writefd, (fd_set*)0, 0)) == -1)
	  printf("Error on Select\n");

	else
	  okayed = 1;
      }while(!okayed);

      if ((rval = write(tty_fd,track,CHUNK)) < 0)
	perror("writing stream message");
      else
	rval_total += rval;

/*      trk = track;

      for (i=0;i<rval;i++)
	{
	  printf("%c",*trk);
	  trk++;
	}*/


      Sent += CHUNK;
      track += CHUNK;
      if (Sent >= len)
	Done = 1;

    }while (!Done);
/*  printf("\n");
  printf("just wrote %d bytes\n",rval_total);*/
  xTrace1(serialp, TR_EVENTS, "SERIAL - Sent %d bytes to serial line",rval_total);
  return TRUE;

}

/*
 *   serialPush
 */
static xmsg_handle_t
serialPush(s, msg)
    XObj s;
    Msg *msg;
{
  int     msg_len;
  msg_len = msgLen(msg);
  xTrace1(serialp, TR_EVENTS, "Serial - about to send % to device",msg_len);

  msgForEach(msg,send_data,(void*)&msg_len);

  xTrace1(serialp, TR_EVENTS, "SERIAL Push - just sent %d bytes",msg_len);

  return XMSG_NULL_HANDLE;
}

/*
 *   serialPop
 */
static xkern_return_t
serialPop(s, ds, dg, h)
    XObj s;
    XObj ds;
    Msg *dg;
    VOID *h;
{

  int k=0;

  xTrace1(serialp, TR_EVENTS, "SERIAL pop, length = %d ", msgLen(dg));
  xAssert(xIsSession(s));

  k = msgLen(dg);
  return xDemux(s, dg);
}


static xkern_return_t
serialOpenDisable(self, hlpRcv, hlpType, p)
    XObj self, hlpRcv, hlpType;
    Part *p;
{
  xTrace0(serialp, TR_EVENTS, "SERIAL open");
  return XK_SUCCESS;
}


static void
getproc_protl(s)
    XObj s;
{
  xTrace0(serialp, TR_EVENTS, "SERIAL getproc_protl");
  xAssert(xIsProtocol(s));
  s->close = serialCloseProtl;
  s->control = serialControlProtl;
  s->open = serialOpen;
  s->openenable = serialOpenEnable;
  s->opendisable = serialOpenDisable;
  s->demux = serialDemux;
}


static void
getproc_sessn(s)
    XObj s;
{
  xTrace0(serialp, TR_EVENTS, "SERIAL getproc_sessn");
  s->push = serialPush;
  s->pop = serialPop;
  s->control = serialControlSessn;
  s->close = serialCloseSessn;
}




