/*
 * $RCSfile: buffer.c,v $
 *
 * Universal Stub Compiler
 *
 * Copyright (c) 1994    Arizona Board of Regents
 *
 * $Log: buffer.c,v $
 * Revision 1.1  1994/10/13 17:43:30  dorgival
 * Initial revision
 *
 */

/*
 * Buffer handling routines
 */

#include "usc.h"
#include "buffer.h"
#include "symbol.h"
#include "codegen_export.h"
#include "gen_utils.h"

extern void usit_add_function (Buffer *source);

void dump_source (void)
{
  cleanbuf (SourceText);
}

void print_source (void)
{
  char *s = BUFFER_STRING (SourceText);
  int i, done = 0;
  if (do_preprocess) {
    usit_add_function (SourceText);
  }
  else if (print_comments) {
    while (*s == '\n')
      s++;
    if (*s == '{') {
      while (*s != '\n' && *s)
        s++;
    }
    while (*s == '\n')
      s++;
    genout("\n/**********************************************************/\n");
    genout  ("/*                                                        */\n");
    genout  ("/* Original Source:                                       */\n");
    genout  ("/*                                                        */\n");
    while (!done) {
      genout("/* ");
      for (i = 0 ; i < 54 && *s && *s != '\n'; s++, i++) {
        genout("%c", *s);
        if (*s == '\t')
          i += 2;
      }
      for ( ; i < 54; i++) 
        genout(" ");
      if (*s == '\n')
        s++;
      if (!(*s))
        done = 1;
      genout(" */\n");
    }
    genout("/*                                                        */\n");
    genout("/**********************************************************/\n\n");
  }
  cleanbuf (SourceText);
}

void catbuf (Buffer *one, Buffer *two)
{
  char *indent, *new;
  int temp = BUFFER_SIZE (one), indentlen;

  assert (BUFFER_LEN (one) == strlen (BUFFER_STRING(one)));
  assert (BUFFER_LEN (two) == strlen (BUFFER_STRING(two)));

  if (BUFFER_STRING(two)[0] == '{')
    indent = make_indent (code_stack_depth - 1);
  else
    indent = make_indent (code_stack_depth);
  indentlen = strmem (indent);

  if (BUFFER_LEN (one) + BUFFER_LEN (two) + indentlen >= BUFFER_SIZE (one)) {
    while (BUFFER_LEN (one) + BUFFER_LEN (two) + indentlen >= 
           BUFFER_SIZE (one)) {
      BUFFER_SIZE (one) *= 2;
    }
    new = (char *)Malloc (BUFFER_SIZE (one) * sizeof (char));
    strcpy (new, BUFFER_STRING (one));
    Free (BUFFER_STRING (one), temp);
    BUFFER_STRING(one) = new;
  }
  if ((one == func_body || one == func_vars) && 
      (BUFFER_STRING(one)[BUFFER_LEN(one) - 1] == '\n' ||
       BUFFER_STRING(one)[0] == '\0')) {
    strcat (BUFFER_STRING (one), indent);
    BUFFER_LEN (one) += indentlen - 1;
  }
  BUFFER_LEN (one) += BUFFER_LEN (two);
  strcat (BUFFER_STRING (one), BUFFER_STRING (two));
  if (strlen (indent))
    Free (indent, strmem (indent));

  assert (BUFFER_LEN (one) == strlen (BUFFER_STRING(one)));

  va_end (args);
}

void writebuf (Buffer *buffer, char *fmt, ...)
{
  va_list args;
  static char string[2056];
  char *new, *indent;
  int stringlen;
  int indentlen;
  int temp = BUFFER_SIZE (buffer);

  va_start (args, fmt);

  assert (BUFFER_LEN (buffer) == strlen (BUFFER_STRING (buffer)));

  vsprintf (string, fmt, args);
  stringlen = strlen (string);
  if (stringlen > 2056)
    crash ("Out of memory: %s", string);
  if (string[0] == '{')
    indent = make_indent (code_stack_depth - 1);
  else
    indent = make_indent (code_stack_depth);
  indentlen = strmem (indent);

  if (BUFFER_LEN (buffer) + stringlen + indentlen >= BUFFER_SIZE (buffer)) {
    while (BUFFER_LEN (buffer) + stringlen + indentlen >= BUFFER_SIZE (buffer))
      BUFFER_SIZE (buffer) *= 2;
    new = (char *)Malloc (BUFFER_SIZE (buffer) * sizeof (char));
    strcpy (new, BUFFER_STRING (buffer));
    Free (BUFFER_STRING (buffer), temp);
    BUFFER_STRING (buffer) = new;
  }
  if ((buffer == func_body || buffer == func_vars) && 
      (BUFFER_STRING(buffer)[0] == '\0' || 
       BUFFER_STRING(buffer)[BUFFER_LEN(buffer) - 1] == '\n')) {
    sprintf (BUFFER_STRING (buffer) + BUFFER_LEN (buffer), "%s", indent);
    BUFFER_LEN (buffer) += indentlen - 1;
  }
  sprintf (BUFFER_STRING (buffer) + BUFFER_LEN (buffer), "%s", string);
  BUFFER_LEN (buffer) += stringlen;
  if (strlen (indent))
    Free (indent, strmem (indent));

  assert (BUFFER_LEN (buffer) == strlen (BUFFER_STRING (buffer)));

  va_end (args);
}

void freebuf (Buffer *buf)
{
  Free (BUFFER_STRING (buf), BUFFER_SIZE(buf));
  Free (buf, sizeof (Buffer));
}

void cleanbuf (Buffer *buffer)
{
  BUFFER_STRING (buffer)[0] = '\0';
  BUFFER_LEN (buffer) = 0;
}

Buffer *makebuf (int size)
{
  Buffer *new = (Buffer *)Malloc (sizeof (Buffer));

  BUFFER_STRING(new) = (char *)Malloc (size * sizeof (char));
  BUFFER_LEN (new) = 0;
  BUFFER_STRING(new)[0] = '\0';
  BUFFER_SIZE(new) = size;
  return new;
}

Buffer *copy_buf (Buffer *buffer)
{
  Buffer *new = (Buffer *)TempMalloc (sizeof (Buffer));
  *new = *buffer;
  BUFFER_STRING (new) = (char *)TempMalloc (BUFFER_SIZE (new));
  strncpy (BUFFER_STRING (new), BUFFER_STRING (buffer), BUFFER_SIZE (new));
  return new;
}

